# **************************************************************************
# *
# * Authors:  David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import os

import pyworkflow.protocol.params as params
from pyworkflow.viewer import DESKTOP_TKINTER, WEB_DJANGO, ProtocolViewer

from pwem.viewers import ChimeraView

from flexutils.protocols.xmipp.protocol_match_and_deform_map_zernike3d import XmippMatchDeformMapZernike3D


class FlexShowMapsViewer(ProtocolViewer):
    """ Show maps using ChimeraX """
    _label = 'show maps ChimeraX'
    _targets = [XmippMatchDeformMapZernike3D]
    _environments = [DESKTOP_TKINTER, WEB_DJANGO]
    OPEN_FILE = "open %s\n"
    VIEW = "view\n"
    VOXEL_SIZE = "volume #%d voxelSize %f\n"

    def _defineParams(self, form):
        form.addSection(label='Show deformation')
        form.addParam('doShowMaps', params.LabelParam,
                      label="Display the maps")

    def _getVisualizeDict(self):
        return {'doShowMaps': self._doShowMaps}

    def _doShowMaps(self, param=None):
        # Get shortest path
        input_vol = self.protocol.input.get()
        reference_vol = self.protocol.reference.get()
        output_vol = self.protocol.deformedMap
        input = os.path.abspath(input_vol.getFileName())
        reference = os.path.abspath(reference_vol.getFileName())
        output = os.path.abspath(output_vol.getFileName())

        scriptFile = self.protocol._getPath('show_maps_chimera.cxc')
        fhCmd = open(scriptFile, 'w')

        fhCmd.write(self.OPEN_FILE % input)
        fhCmd.write(self.OPEN_FILE % reference)
        fhCmd.write(self.OPEN_FILE % output)
        fhCmd.write(self.VOXEL_SIZE % (1, input_vol.getSamplingRate()))
        fhCmd.write(self.VOXEL_SIZE % (2, reference_vol.getSamplingRate()))
        fhCmd.write(self.VOXEL_SIZE % (3, output_vol.getSamplingRate()))

        view = ChimeraView(scriptFile)
        return [view]





import numpy as np
from Bio.PDB.PDBParser import PDBParser
from Bio.PDB.internal_coords import AtomKey

import prody as pd


class PDBUtils:
    # SELECT = {"all": "all", "bb": ["CA", "N", "C"], "ca": ["CA"]}
    SELECT = {"all": None, "bb": "bb", "ca": "ca"}

    def __init__(self, selectionString):
        self.selectionString = self.SELECT[selectionString]
        # self.parser = PDBParser()

    # def parsePDB(self, file):
    #     protein = self.parser.get_structure("protein", file)[0]
    #
    #     # create a selector to filter just the C-alpha atoms from the all atom array
    #     backbone_hedra = []
    #     # debug_backbone_hedra = []  # For debugging purposes
    #     idx_correspondence = dict()
    #     backbone_coords = []
    #     # debug_backbone_coords = []  # For debugging purposes
    #     chain_idx = 1
    #     for chain in protein:
    #         # Compute bond lengths, angles, dihedral angles
    #         chain.atom_to_internal_coordinates(verbose=True)
    #
    #         # Get connection indices only for backbone
    #         if chain.internal_coord.atomArrayIndex:
    #             for hedron in chain.internal_coord.hedra:
    #                 atmNameNdx = AtomKey.fields.atm
    #                 atomArrayIndex = chain.internal_coord.atomArrayIndex
    #                 atomArray = chain.internal_coord.atomArray
    #                 if isinstance(self.selectionString, list):
    #                     select = all([k.akl[atmNameNdx] in self.selectionString for k in hedron])
    #                 elif self.selectionString == "all":
    #                     select = True
    #                 if select:
    #                     new_hedron = []
    #                     for k in hedron:
    #                         hedron_idx = atomArrayIndex.get(k)
    #                         new_key = str(chain_idx) + "_" + str(hedron_idx)
    #                         if not new_key in idx_correspondence.keys():
    #                             idx_correspondence[new_key] = len(idx_correspondence)
    #                             backbone_coords.append(atomArray[hedron_idx][:-1])
    #                         new_hedron.append(idx_correspondence[new_key])
    #                     backbone_hedra.append(new_hedron)
    #                     # debug_backbone_hedra.append([atomArrayIndex.get(k) for k in hedron])
    #
    #             # # Debugging purposes
    #             # atmNameNdx = AtomKey.fields.atm
    #             # atomArrayIndex = chain.internal_coord.atomArrayIndex
    #             # backboneSelect = [
    #             #     atomArrayIndex.get(k) for k in atomArrayIndex.keys() if k.akl[atmNameNdx] in self.selectionString
    #             # ]
    #             # debug_backbone_coords.append(chain.internal_coord.atomArray[backboneSelect])
    #
    #         # Update chain index
    #         chain_idx += 1
    #
    #     backbone_hedra = np.asarray(backbone_hedra, dtype=int)
    #     # debug_backbone_hedra = np.asarray(debug_backbone_hedra, dtype=int)
    #     backbone_coords = np.asarray(backbone_coords, dtype=float)
    #     # debug_backbone_coords = np.vstack(debug_backbone_coords)[:, :-1]
    #
    #     # Debugging purposes
    #     # aux_1 = backbone_coords[np.argsort(np.linalg.norm(backbone_coords, axis=1))]
    #     # aux_2 = debug_backbone_coords[np.argsort(np.linalg.norm(debug_backbone_coords, axis=1))]
    #     # same_coords = np.array_equal(aux_1[:, 0], aux_2[:, 0]) and \
    #     #               np.array_equal(aux_1[:, 1], aux_2[:, 1]) and \
    #     #               np.array_equal(aux_1[:, 2], aux_2[:, 2])
    #
    #     return backbone_coords, backbone_hedra

    @classmethod
    def find_dihedrals(cls, bond_indices):
        # Convert bond indices to a set for faster lookup
        bonds = set(map(tuple, map(sorted, bond_indices)))
        dihedrals = []
        # Create a dictionary to store neighbors for each atom
        neighbors = {}
        for bond in bonds:
            atom1, atom2 = bond
            if atom1 not in neighbors:
                neighbors[atom1] = set()
            if atom2 not in neighbors:
                neighbors[atom2] = set()
            neighbors[atom1].add(atom2)
            neighbors[atom2].add(atom1)
        for bond in bonds:
            atom1, atom2 = bond
            # Potential neighbors for the dihedral
            for neighbor1 in neighbors[atom1]:
                if neighbor1 == atom2:
                    continue
                for neighbor2 in neighbors[atom2]:
                    if neighbor2 == atom1 or neighbor2 == neighbor1:
                        continue
                    # Check for unique dihedral indices
                    dihedral = tuple(sorted([neighbor1, atom1, atom2, neighbor2]))
                    if dihedral not in dihedrals:
                        dihedrals.append(dihedral)
        return np.array(dihedrals)
    def parsePDB(self, file):
        # Parse model file based on subset
        structure = pd.parsePDB(file, subset=self.selectionString, compressed=False)

        # Infer bonds from parsed model
        _ = structure.inferBonds()

        # Get bond indices
        bonds = np.asarray([bond.getIndices() for bond in structure.iterBonds()])

        # Get dihedral angles
        dihedrals = PDBUtils.find_dihedrals(bonds)

        # Get atom coordinates
        atom_coordinates = structure.getCoords()

        # Alpha carbon indices
        ca_indices = structure.select("name CA").getIndices()

        return atom_coordinates, ca_indices, dihedrals


    @staticmethod
    def calc_bond(coords, connectivity):
        def ReLU(x):
            return x * (x > 0)

        px = coords[connectivity[:, 0]]
        py = coords[connectivity[:, 1]]
        pz = coords[connectivity[:, 2]]
        dst_1 = np.sqrt(ReLU(np.sum((px - py) ** 2, axis=1)))
        dst_2 = np.sqrt(ReLU(np.sum((py - pz) ** 2, axis=1)))

        return np.stack([dst_1, dst_2]).T

    @staticmethod
    def calc_angle(coords, connectivity, degrees=True):
        p0 = coords[connectivity[:, 0]]
        p1 = coords[connectivity[:, 1]]
        p2 = coords[connectivity[:, 2]]
        b0 = p0 - p1
        b1 = p2 - p1
        ang = np.sum(b0 * b1, axis=1)
        n0 = np.linalg.norm(b0, axis=1) * np.linalg.norm(b1, axis=1)
        ang = np.nan_to_num(ang / n0)
        # ang = np.min(np.max(ang, axis=1), axis=0)
        ang = np.arccos(ang)

        if degrees:
            ang *= 180 / np.pi

        return ang



# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import os
import numpy as np
from xmipp_metadata.image_handler import ImageHandler

from pyworkflow import BETA
from pyworkflow.protocol.params import PointerParam, IntParam, MultiPointerParam
import pyworkflow.utils as pwutils
from pyworkflow.utils.properties import Message
from pyworkflow.gui.dialog import askYesNo
from pyworkflow.object import Boolean, Integer

from pwem.protocols import ProtAnalysis3D, ProtFlexBase
from pwem.objects import SetOfVolumesFlex, VolumeFlex

import flexutils
from flexutils.utils import getOutputSuffix, computeNormRows
import flexutils.constants as const

import xmipp3


class ProtFlexClusterSpace(ProtAnalysis3D, ProtFlexBase):
    """ Interactive clustering of conformational spaces """

    _label = 'cluster space'
    _devStatus = BETA
    OUTPUT_PREFIX = 'flexible3DClasses'

    # --------------------------- DEFINE param functions ----------------------
    def _defineParams(self, form):
        form.addSection(label='General parameters')
        form.addParam('particles', PointerParam, label="Particles to annotate",
                      pointerClass='SetOfParticlesFlex', important=True,
                      help="Particles must have a flexibility information associated (Zernike3D, CryoDrgn...")
        form.addHidden('volumes', PointerParam, label="Priors", allowsNull=True,
                       pointerClass="SetOfVolumesFlex, VolumeFlex",
                       condition="particles and particles.getFlexInfo().getProgName() == 'Zernike3D'",
                       help='A set of volumes with Zernike3D coefficients associated (computed using '
                            '"Reference map" as reference) to add as prior information to the Zernike3D '
                            'space')
        form.addParam('priors', MultiPointerParam, label="Priors", allowsNull=True,
                      pointerClass="SetOfVolumes, Volume",
                      condition="particles and particles.getFlexInfo().getProgName() == 'Zernike3D'",
                      help='Volumes with Zernike3D coefficients associated (computed using '
                           '"Refernce map" as reference) to add as prior information to the Zernike3D '
                           'space')
        form.addParam('boxSize', IntParam, label="Box size",
                      condition="particles and particles.getFlexInfo().getProgName() == 'CryoDRGN'",
                      help="Volumes generated from the CryoDrgn network will be resampled to the "
                           "chosen box size (only for the visualization).")

    # --------------------------- INSERT steps functions ----------------------
    def _insertAllSteps(self):
        self._insertFunctionStep(self.launchInteractiveClustering, interactive=True)

    def _createOutput(self):
        particles = self.particles.get()
        partIds = list(particles.getIdSet())
        sr = particles.getSamplingRate()
        progName = particles.getFlexInfo().getProgName()

        # Get right imports
        if progName == const.NMA:
            createFn = self._createSetOfClassesStructFlex
            from pwem.objects import ClassStructFlex as Class
            from pwem.objects import AtomStructFlex as Rep
            from pwem.objects import SetOfClassesStructFlex as SetOfClasses
        else:
            createFn = self._createSetOfClassesFlex
            from pwem.objects import ClassFlex as Class
            from pwem.objects import VolumeFlex as Rep
            from pwem.objects import SetOfClassesFlex as SetOfClasses

        # Read KMean coefficients
        z_space_vw = []
        with open(self._getExtraPath('saved_selections.txt')) as f:
            lines = f.readlines()
            for line in lines:
                z_space_vw.append(np.fromstring(line, dtype=float, sep=' '))
        z_space_vw = np.asarray(z_space_vw[self.num_vol:])

        # Read KMean labels
        km_labels = np.loadtxt(self._getExtraPath('kmean_labels.txt'), delimiter=" ")
        if self.num_vol > 0:
            km_labels = km_labels[:-self.num_vol]

        # Create SetOfClasses3D
        suffix = getOutputSuffix(self, SetOfClasses)
        flexClasses = createFn(particles, suffix, progName=progName)

        # Popoulate SetOfClasses3D with KMean particles
        for clInx in range(z_space_vw.shape[0]):
            currIds = np.where(km_labels == clInx)[0]

            newClass = Class()
            newClass.copyInfo(particles)
            newClass.setHasCTF(particles.hasCTF())
            newClass.setAcquisition(particles.getAcquisition())
            representative = Rep(progName=progName)

            # ****** Fill representative information *******
            if particles.getFlexInfo().getProgName() == const.ZERNIKE3D:
                reference = particles.getFlexInfo().refMap.get()

                # Resize coefficients
                factor = (ImageHandler().read(reference).getDimensions()[0] / 64)
                z_space_vw[clInx] *= factor

                representative.setLocation(reference)

            elif particles.getFlexInfo().getProgName() == const.CRYODRGN:
                from cryodrgn.utils import generateVolumes
                generateVolumes(z_space_vw[clInx], particles.getFlexInfo()._cryodrgnWeights.get(),
                                particles.getFlexInfo()._cryodrgnConfig.get(), self._getExtraPath(),
                                downsample=self.boxSize.get(), apix=particles.getSamplingRate())
                ImageHandler().scaleSplines(self._getExtraPath('vol_000.mrc'),
                                            self._getExtraPath('class_%d.mrc') % clInx,
                                            finalDimension=particles.getXDim(), overwrite=True)
                representative.setLocation(self._getExtraPath('class_%d.mrc') % clInx)

            elif particles.getFlexInfo().getProgName() == const.HETSIREN:
                from flexutils.utils import generateVolumesHetSIREN
                generateVolumesHetSIREN(particles.getFlexInfo().modelPath.get(), z_space_vw[clInx],
                                        self._getExtraPath(), step=particles.getFlexInfo().coordStep.get())
                ImageHandler().scaleSplines(self._getExtraPath('decoded_map_class_01.mrc'),
                                            self._getExtraPath('class_%d.mrc') % clInx,
                                            finalDimension=particles.getXDim(), overwrite=True)
                representative.setLocation(self._getExtraPath('class_%d.mrc') % clInx)

            elif particles.getFlexInfo().getProgName() == const.NMA:
                reference = particles.getFlexInfo().refStruct.get()
                subset = particles.getFlexInfo().atomSubset

                representative.getFlexInfo().atomSubset = subset
                representative.setLocation(reference)

            representative.setZFlex(z_space_vw[clInx])
            representative.getFlexInfo().copyInfo(particles.getFlexInfo())
            # ********************

            if hasattr(representative, "setSamplingRate"):
                representative.setSamplingRate(sr)
            newClass.setRepresentative(representative)

            flexClasses.append(newClass)

            enabledClass = flexClasses[newClass.getObjId()]
            enabledClass.enableAppend()
            for itemId in currIds:
                item = particles[partIds[itemId]]
                item._xmipp_subtomo_labels = Integer(clInx + 1)
                enabledClass.append(item)

            flexClasses.update(enabledClass)

        # Save new output
        name = self.OUTPUT_PREFIX + suffix
        args = {}
        args[name] = flexClasses
        self._defineOutputs(**args)
        self._defineSourceRelation(particles, flexClasses)

    # --------------------------- STEPS functions -----------------------------
    def launchInteractiveClustering(self):
        particles = self.particles.get()
        self.num_vol = 0


        # ********* Get Z space *********
        z_space = []
        for particle in particles.iterItems():
            z_space.append(particle.getZFlex())
        z_space = np.asarray(z_space)

        if particles.getFlexInfo().getProgName() == const.ZERNIKE3D:
            reference = particles.getFlexInfo().refMap.get()
            mask = particles.getFlexInfo().refMask.get()
            volumes = self.priors

            # Copy original reference and mask to extra
            ih = ImageHandler()
            ih.convert(reference, self._getExtraPath("reference_original.mrc"), overwrite=True)
            ih.convert(mask, self._getExtraPath("mask_reference_original.mrc"), overwrite=True)

            # Resize reference map to increase real time conformation inspection performance
            inputFile = reference
            outFile = self._getExtraPath('reference.mrc')
            if not os.path.isfile(outFile):
                if pwutils.getExt(inputFile) == ".mrc":
                    inputFile += ":mrc"
                self.runJob("xmipp_image_resize",
                            "-i %s -o %s --dim %d " % (inputFile,
                                                       outFile,
                                                       64), numberOfMpi=1, env=xmipp3.Plugin.getEnviron())

            # Resize mask
            inputFile = mask
            outFile = self._getExtraPath('mask.mrc')
            if not os.path.isfile(outFile):
                if pwutils.getExt(inputFile) == ".mrc":
                    inputFile += ":mrc"
                self.runJob("xmipp_image_resize",
                            "-i %s -o %s --dim %d " % (inputFile,
                                                       outFile,
                                                       64), numberOfMpi=1, env=xmipp3.Plugin.getEnviron())
                self.runJob("xmipp_transform_threshold",
                            "-i %s -o %s --select below 0.01 "
                            "--substitute binarize " % (outFile, outFile), numberOfMpi=1,
                            env=xmipp3.Plugin.getEnviron())

            z_space_vol = []
            if volumes:
                for pointer in volumes:
                    item = pointer.get()
                    if isinstance(item, VolumeFlex):
                        z_space_vol.append(item.getZFlex())
                    elif isinstance(item, SetOfVolumesFlex):
                        for volume in item.iterItems():
                            z_space_vol.append(volume.getZFlex())
            z_space_vol = np.asarray(z_space_vol)

            # Get useful parameters
            self.num_vol = z_space_vol.shape[0]
            if self.num_vol > 0:
                z_space = np.vstack([z_space, z_space_vol])

            # Resize coefficients
            z_space = (64 / ImageHandler().read(reference).getDimensions()[0]) * z_space
        # ********************

        # Generate files to call command line
        file_z_space = self._getExtraPath("z_space.txt")
        file_interp_val = self._getExtraPath("interp_val.txt")
        np.savetxt(file_z_space, z_space)

        # Compute/Read UMAP or PCA
        file_coords = self._getExtraPath("red_coords.txt")
        red_space = []
        for particle in particles.iterItems():
            red_space.append(particle.getZRed())
        red_space = np.asarray(red_space)
        np.savetxt(file_coords, red_space)

        # ********* Get interpolation value for coloring the space *********
        if particles.getFlexInfo().getProgName() == const.ZERNIKE3D:
            interp_val = computeNormRows(z_space)
        else:
            interp_val = np.ones([1, z_space.shape[0]])
        # *********

        # Generate files to call command line
        if not os.path.isfile(file_interp_val):
            np.savetxt(file_interp_val, interp_val)
        path = self._getExtraPath()

        # ********* Run viewer *********
        needsPackages = None

        if particles.getFlexInfo().getProgName() == const.ZERNIKE3D:
            L1 = particles.getFlexInfo().L1.get()
            L2 = particles.getFlexInfo().L2.get()
            args = "--data %s --z_space %s --interp_val %s --path %s " \
                   "--L1 %d --L2 %d --n_vol %d --mode Zernike3D" \
                   % (file_coords, file_z_space, file_interp_val, path, L1, L2,
                      self.num_vol)

        elif particles.getFlexInfo().getProgName() == const.CRYODRGN:
            needsPackages = [const.CRYODRGN, ]
            args = "--data %s --z_space %s --interp_val %s --path %s " \
                   "--weights %s --config %s --boxsize %d --sr %f --mode CryoDrgn" \
                   % (file_coords, file_z_space, file_interp_val, path,
                      particles.getFlexInfo()._cryodrgnWeights.get(),
                      particles.getFlexInfo()._cryodrgnConfig.get(), self.boxSize.get(),
                      particles.getSamplingRate())

        elif particles.getFlexInfo().getProgName() == const.HETSIREN:
            args = "--data %s --z_space %s --interp_val %s --path %s " \
                   "--weights %s --step %d --sr %f --mode HetSIREN" \
                   % (file_coords, file_z_space, file_interp_val, path,
                      particles.getFlexInfo().modelPath.get(),
                      particles.getFlexInfo().coordStep.get(),
                      particles.getSamplingRate())

        elif particles.getFlexInfo().getProgName() == const.NMA:
            args = "--data %s --z_space %s --interp_val %s --path %s " \
                   "--weights %s --sr %f --boxsize %d --mode NMA" \
                   % (file_coords, file_z_space, file_interp_val, path,
                      particles.getFlexInfo().modelPath.get(),
                      particles.getSamplingRate(), particles.getXDim())

        program = "viewer_3d_pc.py"
        program = flexutils.Plugin.getProgram(program, needsPackages=needsPackages)
        self.runJob(program, args)

        # *********

        if os.path.isfile(self._getExtraPath("saved_selections.txt")) and \
           askYesNo(Message.TITLE_SAVE_OUTPUT, Message.LABEL_SAVE_OUTPUT, None):
            self._createOutput()

    # --------------------------- INFO functions -----------------------------
    def _summary(self):
        summary = []
        if self.getOutputsSize() >= 1:
            for _, outClasses in self.iterOutputAttributes():
                summary.append("Output *%s*:" % outClasses.getNameId().split('.')[1])
                summary.append("    * Total clustered classes: *%s*" % outClasses.getSize())
        else:
            summary.append("Output clustered classes not ready yet")
        return summary

    def _methods(self):
        return [
            "Interactive automatic clustering of conformational spaces",
        ]

    # ----------------------- VALIDATE functions ----------------------------------------
    def validate(self):
        """ Try to find errors on define params. """
        errors = []
        particles = self.particles.get()
        if particles.getFirstItem().getZRed().size == 0:
            errors.append("Particles do not have any dimensionality reduced version of the flexibility information "
                          "they store. Please, used the dimensionality reduction protocol available "
                          "in Flexutils Plugin to generate a valid set of particles")
        elif particles.getFirstItem().getZRed().size != 3:
            errors.append("Cluster space is only implemented for reduced spaces of 3 dimensions. Please, used the "
                          "dimensionality reduction protocol available in Flexutils Plugin to generate a valid "
                          "set of particles")

        # Check CryoDRGN boxsize parameter is set as it is mandatory
        if particles.getFlexInfo().getProgName() == 'CryoDRGN' and self.boxSize.get() is None:
            errors.append("Boxsize parameter needs to be set to an integer value smaller than or equal "
                          "to the boxsize used internally to train the CryoDRGN network")

        return errors



# **************************************************************************
# *
# * Authors:  David Herreros (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import os
from pathos.multiprocessing import ProcessingPool as Pool
import subprocess

from pyworkflow.viewer import Viewer
from pyworkflow.utils.process import buildRunCommand

from flexutils.protocols.xmipp.protocol_angular_align_zernike3deep import TensorflowProtAngularAlignmentZernike3Deep
from flexutils.protocols.xmipp.protocol_angular_align_het_siren import TensorflowProtAngularAlignmentHetSiren
from flexutils.protocols.xmipp.protocol_angular_align_deep_pose import TensorflowProtAngularAlignmentDeepPose
from flexutils.protocols.xmipp.protocol_angular_align_homo_siren import TensorflowProtAngularAlignmentHomoSiren

import flexutils.constants as const
import flexutils


class TensorboardViewer(Viewer):
    """ Tensorboard visualization of neural networks """
    _label = 'viewer conformational landscape'
    _targets = [TensorflowProtAngularAlignmentZernike3Deep,
                TensorflowProtAngularAlignmentHetSiren,
                TensorflowProtAngularAlignmentDeepPose,
                TensorflowProtAngularAlignmentHomoSiren]

    def __init__(self, **kwargs):
        Viewer.__init__(self, **kwargs)
        self._data = None

    def _visualize(self, obj, **kwargs):
        logdir_path = self.protocol._getExtraPath(os.path.join("network", "logs"))

        def launchViewerNonBlocking(args):
            (logdir_path, ) = args

            if not os.path.isdir(logdir_path):
                raise NotADirectoryError("Logs directory has not been generated for this network. "
                                         "Please, execute the training again with an updated version "
                                         "of the Flexutils-Tensorflow toolkit to generate the logs.")

            # Run tensorboard
            args = f"--logdir_path {logdir_path}"
            program = os.path.join(const.VIEWERS, "tensorboard", "setup_tensorboard.py")
            program = flexutils.Plugin.getProgram(program)

            command = buildRunCommand(program, args, 1)
            subprocess.Popen(command, shell=True)

        # Launch with Pathos
        p = Pool()
        # p.restart()
        p.apipe(launchViewerNonBlocking, args=(logdir_path, ))
        # p.join()
        # p.close()

        return []


# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import os

import importlib

import pyworkflow.plugin as pwplugin
import pyworkflow.utils as pwutils

from pwem import Config as emConfig

import flexutils
from flexutils.constants import CONDA_YML


__version__ = "3.1.6"
_logo = "icon.png"
_references = []


class Plugin(pwplugin.Plugin):

    @classmethod
    def getEnvActivation(cls):
        return "conda activate flexutils"

    @classmethod
    def getTensorflowActivation(cls):
        return "conda activate flexutils-tensorflow"

    @classmethod
    def getProgram(cls, program, python=False, cuda=False, chimera=False, needsPackages=None):
        """ Return the program binary that will be used. """
        scipion_packages = []
        env_variables = ""
        if needsPackages is not None:
            for package_name in needsPackages:
                package_name = package_name.lower()
                try:
                    package = importlib.import_module(package_name)
                    package_loc = importlib.util.find_spec(package_name).submodule_search_locations[0]
                    scipion_packages.append(os.path.dirname(package_loc))
                except ImportError:
                    raise ImportError(f"Package {package_name} is not installed in Scipion")
                package_env_vars = package.Plugin.getVars()
                for item, value in package_env_vars.items():
                    if package_name.lower() in item.lower():
                        env_variables += " {}='{}'".format(item, value)
        scipion_packages = ":".join(scipion_packages)
        cmd = '%s %s && ' % (cls.getCondaActivationCmd(), cls.getEnvActivation())

        if cuda:
            cmd += 'LD_LIBRARY_PATH=$CONDA_PREFIX/lib/:$LD_LIBRARY_PATH '

        if chimera:
            with pwutils.weakImport("chimera"):
                from chimera import Plugin as chimeraPlugin
                cmd += "CHIMERA_HOME=%s " % chimeraPlugin.getHome()

        if python:
            if needsPackages is not None:
                cmd += "PYTHONPATH=$PYTHONPATH:%s %s python " % (scipion_packages, env_variables)
            else:
                cmd += "python "
        return cmd + '%(program)s ' % locals()

    @classmethod
    def getTensorflowProgram(cls, program, python=True, log_level=2):
        cmd = '%s %s && ' % (cls.getCondaActivationCmd(), cls.getTensorflowActivation())
        if python:
            # import pyworkflow, pwem, xmipp3
            # pyworkflow_path = os.path.join(pyworkflow.__path__[0], "..")
            # pywem_path = os.path.join(pwem.__path__[0], "..")
            # xmipp3_path = os.path.join(xmipp3.__path__[0], "..")
            # paths = [os.path.join(flexutils.__path__[0], ".."), pyworkflow_path, pywem_path, xmipp3_path]
            return cmd + ' TF_CPP_MIN_LOG_LEVEL=%(log_level)d python %(program)s ' % locals()
        else:
            return cmd + ' TF_CPP_MIN_LOG_LEVEL=%(log_level)d %(program)s ' % locals()

    @classmethod
    def getCommand(cls, program, args, python=True):
        return cls.getProgram(program, python) + args

    def defineBinaries(cls, env):
        def getCondaInstallationFlexutils():
            installationCmd = f'if [ $(basename "$PWD") = flexutils-{__version__} ]; then cd ..; fi && '
            installationCmd += "git clone -b devel https://github.com/I2PC/Flexutils-Scripts.git && "
            installationCmd += "cd Flexutils-Scripts && "
            installationCmd += "bash install.sh && touch flexutils_installed && cd .."
            return installationCmd

        def getCondaInstallationTensorflow():
            conda_init = cls.getCondaActivationCmd()
            branch = "devel" if cls.inDevelMode() else "master"
            installationCmd = f'if [ $(basename "$PWD") = flexutils-{__version__} ]; then cd ..; fi && '
            installationCmd += f"{conda_init} conda activate flexutils && " \
                               f' if [ ! -d "Flexutils-Toolkit" ]; then git clone -b {branch} https://github.com/I2PC/Flexutils-Toolkit.git; fi && ' \
                               f"cd Flexutils-Toolkit && " \
                               f"bash install.sh && touch flexutils_tensorflow_installed && cd .."
            return installationCmd

        def getUpdateCommands():
            conda_init = cls.getCondaActivationCmd()
            updateCmd = f'if [ $(basename "$PWD") = flexutils-{__version__} ]; then cd ..; fi && '
            updateCmd += f"{conda_init} conda activate flexutils && "
            updateCmd += "echo '###### Updating scripts.... ######' && "
            updateCmd += "cd Flexutils-Scripts && "
            updateCmd += "git pull && "
            updateCmd += "pip install -e . && "
            updateCmd += "cd .. && "
            updateCmd += "echo '###### Script updated succesfully! ######' && "

            updateCmd += "echo '###### Updating NN binaries.... ######' && "
            updateCmd += "cd Flexutils-Toolkit && "
            updateCmd += "git pull && "
            updateCmd += "bash install.sh && "
            updateCmd += "cd .. && "
            updateCmd += "echo '###### Binaries updated succesfully! ######' && "
            updateCmd += f"cd flexutils-{__version__} && "
            updateCmd += "touch flexutils_updated"
            return updateCmd

        binary_path = os.path.join(emConfig.EM_ROOT, f'flexutils-{__version__}')
        commands = []

        if not os.path.isfile(os.path.join(binary_path, os.path.join("..", "Flexutils-Scripts", "flexutils_installed"))):
            installationEnv = getCondaInstallationFlexutils()
            commands.append((installationEnv, [os.path.join("..", "Flexutils-Scripts", "flexutils_installed")]))

        if not os.path.isfile(os.path.join(binary_path, os.path.join("..", "Flexutils-Toolkit", "flexutils_tensorflow_installed"))):
            installationTensorflow = getCondaInstallationTensorflow()
            commands.append((installationTensorflow, [os.path.join("..", "Flexutils-Toolkit", "flexutils_tensorflow_installed")]))

        if os.path.isfile(os.path.join(binary_path, "flexutils_tensorflow_updated")):
            os.remove(os.path.join(binary_path, "flexutils_tensorflow_updated"))
        commands.append((getUpdateCommands(), ["flexutils_updated"]))

        env.addPackage('flexutils', version=__version__,
                       commands=commands,
                       tar="void.tgz",
                       default=True)



# **************************************************************************
# *
# * Authors:     David Herreros (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import os
import glob
import numpy as np
from scipy import stats
import re
from xmipp_metadata.image_handler import ImageHandler

import pyworkflow.protocol.params as params
import pyworkflow.utils as pwutils
# import pyworkflow.protocol.constants as cons

from pwem import emlib
from pwem.objects import Volume
from pwem.protocols import ProtReconstruct3D
from pwem.objects import ParticleFlex, SetOfParticlesFlex

from xmipp3.convert import writeSetOfImages, imageToRow, coordinateToRow
# from xmipp3.base import isXmippCudaPresent
import xmipp3

import flexutils
import flexutils.constants as const
from flexutils.utils import getXmippFileName


class XmippProtReconstructZART(ProtReconstruct3D):
    """    
    Reconstruct a volume using ZART algorithm from a given SetOfParticles.
    """
    _label = 'reconstruct ZART'

    def __init__(self, **args):
        ProtReconstruct3D.__init__(self, **args)
        self.stepsExecutionMode = params.STEPS_PARALLEL

    #--------------------------- DEFINE param functions --------------------------------------------   
    def _defineParams(self, form):
        form.addSection(label='Input')

        form.addHidden(params.USE_GPU, params.BooleanParam, default=False,
                       label="Use GPU for execution",
                       help="This protocol has both CPU and GPU implementation.\
                       Select the one you want to use.")

        form.addHidden(params.GPU_LIST, params.StringParam, default='0',
                       expertLevel=params.LEVEL_ADVANCED,
                       label="Choose GPU IDs",
                       help="Add a list of GPU devices that can be used")

        form.addParam('inputParticles', params.PointerParam, pointerClass='SetOfParticles, SetOfParticlesFlex',
                      pointerCondition='hasAlignmentProj',
                      label="Input particles",  
                      help='Select the input images from the project.')
        form.addParam('ctfCorrected', params.BooleanParam, default=False,
                      label="Are particles CTF corrected?",
                      help="If particles are not CTF corrected, set to 'No' to perform "
                           "a Weiner filter based corerction")
        form.addParam('initialMap', params.PointerParam, pointerClass='Volume',
                      label="Initial map",
                      allowsNull=True,
                      help='If provided, this map will be used as the initialization of the reconstruction '
                           'process. Otherwise, an empty volume will be used')
        form.addParam('useZernike', params.BooleanParam, default=False,
                      condition="inputParticles and isinstance(inputParticles, SetOfParticlesFlex)",
                      label="Correct motion blurred artifacts?",
                      help="Correct the conformation of the particles during the reconstruct process "
                           "to reduce motion blurred artifacts and increase resolution. Note that this "
                           "option requires that the particles have a set of Zernike3D coefficients associated. "
                           "Otherwise, the parameter should be set to 'No'")
        form.addParam('recMask', params.PointerParam, pointerClass='VolumeMask',
                      allowsNull=True,
                      expertLevel=params.LEVEL_ADVANCED,
                      label="Reconstruction mask",
                      help="Mask used to restrict the reconstruction space to increase performance.")
        form.addParam('niter', params.IntParam, default=13,
                      label="Number of ZART iterations to perform",
                      help="In general, the bigger the number the sharper the volume. We recommend "
                           "to run at least 8 iteration for better results")
        form.addParam('reg', params.FloatParam, default=1e-5, expertLevel=params.LEVEL_ADVANCED,
                      label='ART lambda',
                      help="This parameter determines how fast ZART will converge to the reconstruction. "
                           "Note that larger values may lead to divergence.")
        form.addParam('dThr', params.FloatParam, default=1e-6,
                      label="Denoising threshold",
                      help="Larger values will decrease the noise levels more efficiently, although protein signal "
                           "might suffer unwanted modifications if the value is too large.")
        form.addParam('save_pr', params.BooleanParam, default=False, expertLevel=params.LEVEL_ADVANCED,
                      label="Save partial reconstructions for every ZART iteration?")
        form.addParam('onlyPositive', params.BooleanParam, default=False, expertLevel=params.LEVEL_ADVANCED,
                      label="Remove negative values from reconstructed volume?")
        form.addParam('mode', params.EnumParam, choices=['Reconstruct', 'Gold standard', 'Multiresolution'],
                      default=0, display=params.EnumParam.DISPLAY_HLIST,
                      label="Reconstruction mode",
                      help="\t * Reconstruct: usual reconstruction of a single volume using all the images "
                           "in the dataset\n"
                           "\t * Gold standard: volumes halves reconstruction for FSC and local resolution "
                           "computations\n"
                           "\t * Multiresolution: local resolution analysis during reconstruction to determine "
                           "which areas of the map can be improved further with the ZART algorithm")
        form.addParam('levels', params.IntParam, default=3, condition="mode==2",
                      label="Number of multiresolution levels")

        form.addParallelSection(threads=4, mpi=1)

    #--------------------------- INSERT steps functions --------------------------------------------
    def _insertAllSteps(self):
        depsConvert = []
        depsReconstruct = []
        convert = self._insertFunctionStep(self.convertInputStep, prerequisites=[])
        depsConvert.append(convert)
        refFile = self.initialMap.get().getFileName() if self.initialMap.get() else None
        recMask = self.recMask.get().getFileName() if self.recMask.get() else None
        if self.mode.get() == 0:
            particlesMd = self._getTmpPath('corrected_particles.xmd')
            reconstruct = self._insertFunctionStep(self.reconstructStep, particlesMd,
                                                   "final_reconstruction.mrc", refFile, 2,
                                                   self.niter.get(), recMask,
                                                   prerequisites=depsConvert)
            depsReconstruct.append(reconstruct)
        elif self.mode.get() == 1:
            particlesHalvesMd = [self._getTmpPath('corrected_particles_half_000001.xmd'),
                                 self._getTmpPath('corrected_particles_half_000002.xmd')]
            for idx, fileMd in enumerate(particlesHalvesMd):
                outFile = "final_reconstruction_%d.mrc" % (idx + 1)
                reconstruct = self._insertFunctionStep(self.reconstructStep, fileMd, outFile, refFile,
                                                       2, self.niter.get(), recMask,
                                                       prerequisites=depsConvert)
                depsReconstruct.append(reconstruct)
        elif self.mode.get() == 2:
            depsMask = []
            current_iter = 0
            niter = self.niter.get()
            level = 1
            particlesHalvesMd = [self._getTmpPath('corrected_particles_half_000001.xmd'),
                                 self._getTmpPath('corrected_particles_half_000002.xmd')]

            for idx, fileMd in enumerate(particlesHalvesMd):
                outFile = "final_reconstruction_%d_level_%d.mrc" % (idx + 1, level)
                reconstruct = self._insertFunctionStep(self.reconstructStep, fileMd, outFile, refFile,
                                                       2, 2, recMask,
                                                       prerequisites=depsConvert)
                depsReconstruct.append(reconstruct)

            resMask = self._insertFunctionStep(self.resolutionMaskStep, level,
                                               prerequisites=depsReconstruct)
            depsMask.append(resMask)

            resMaskFile = self._getTmpPath("resMask.mrc")
            while current_iter < niter:
                depsReconstruct = []
                level += 1
                current_iter += 2
                for idx, fileMd in enumerate(particlesHalvesMd):
                    outFile = "final_reconstruction_%d_level_%d.mrc" % (idx + 1, level)
                    refFile = self._getExtraPath("final_reconstruction_%d_level_%d.mrc" % (idx + 1, level - 1))
                    reconstruct = self._insertFunctionStep(self.reconstructStep, fileMd, outFile, refFile,
                                                           1, 2, resMaskFile,
                                                           prerequisites=depsMask)
                    depsReconstruct.append(reconstruct)

                depsMask = []
                resMask = self._insertFunctionStep(self.resolutionMaskStep, level,
                                                   prerequisites=depsReconstruct)
                depsMask.append(resMask)

        self._insertFunctionStep(self.createOutputStep, prerequisites=depsReconstruct)

        if not self.save_pr.get():
            pwutils.cleanPattern(self._getExtraPath("*_iter_*.mrc"))
            pwutils.cleanPattern(self._getExtraPath("meanMap.mrc"))
            pwutils.cleanPattern(self._getExtraPath("monoresResolutionChimera.mrc"))
            pwutils.cleanPattern(self._getExtraPath("monoresResolutionMap.mrc"))
            pwutils.cleanPattern(self._getExtraPath("refinedMask.mrc"))
        
    def reconstructStep(self, inputMd, outFile, refFile, step, niter, mask):
        params = self.defineZARTArgs(inputMd, outFile, niter, step, mask)

        if refFile:
            params += " --ref %s" % getXmippFileName(refFile)

        if self.usesGpu():
            params += " --debug_iter"
            env = xmipp3.Plugin.getEnviron()
            env["CUDA_VISIBLE_DEVICES"] = ','.join([str(elem) for elem in self.getGpuList()])
            self.runJob('xmipp_cuda_forward_art_zernike3d', params, env=env)
        else:
            if self.numberOfThreads.get() == 1:
                self.runJob('xmipp_forward_art_zernike3d', params, numberOfMpi=1, env=xmipp3.Plugin.getEnviron())
            else:
                params += " --thr %d" % self.numberOfThreads.get()
                self.runJob('xmipp_parallel_forward_art_zernike3d', params, numberOfMpi=1, env=xmipp3.Plugin.getEnviron())

        # if self.useGpu.get():
        #     if self.numberOfMpi.get()>1:
        #         self.runJob('xmipp_cuda_reconstruct_fourier', params, numberOfMpi=len((self.gpuList.get()).split(','))+1)
        #     else:
        #         self.runJob('xmipp_cuda_reconstruct_fourier', params)
        # else:
        #     if self.legacy.get():
        #         self.runJob('xmipp_reconstruct_fourier', params)
        #     else:
        #         self.runJob('xmipp_reconstruct_fourier_accel', params)

    def resolutionMaskStep(self, level):
        half_1 = self._getExtraPath("final_reconstruction_1_level_%d.mrc" % level)
        half_2 = self._getExtraPath("final_reconstruction_2_level_%d.mrc" % level)
        sr = self.inputParticles.get().getSamplingRate()
        low = 2*sr
        high = 2*sr*(self.levels.get() + 1)

        params = "--vol %s --vol2 %s --sampling_rate %f --noiseonlyinhalves " \
                 "--minRes %f --maxRes %f --step 0.5 -o %s --significance 0.95 --threads %d" % \
                 (half_1, half_2, sr, low, high, self._getExtraPath(), self.numberOfThreads.get())

        # if mask:
        #     params += ' --mask %s' % mask
        # else:
        dim = self.inputParticles.get().getXDim()
        mask = self._getTmpPath("sphere_mask.mrc")
        r = int(0.5 * dim)
        ImageHandler().write(np.zeros([dim, dim, dim]), mask, overwrite=True)
        mask_params = "-i %s --mask circular -%d --create_mask %s" % (mask, r, mask)
        self.runJob('xmipp_transform_mask', mask_params, numberOfMpi=1, env=xmipp3.Plugin.getEnviron())
        params += ' --mask %s' % mask

        self.runJob('xmipp_resolution_monogenic_signal', params, numberOfMpi=1, env=xmipp3.Plugin.getEnviron())

        resMap = ImageHandler().read(self._getExtraPath("monoresResolutionMap.mrc")).getData()
        oriMask = ImageHandler().read(mask).getData()

        # _, edges = np.histogram(resMap, bins=3, range=[2*sr, 12.0])

        levels = self.levels.get()
        values = resMap.flatten()
        edges = stats.mstats.mquantiles(values[values != 0.0], np.linspace(0.0, 1.0, num=levels+1))
        mid_points = 0.5 * (edges[1:] + edges[:-1])
        sigmas = mid_points / (2*sr)
        # sigmas *= np.sqrt(1 / (2 * np.log(2)))
        # sigmas *= np.sqrt(np.log(2) / (2 * np.pi * np.pi))
        sigmas = np.floor(sigmas).astype(int)
        # sigmas[sigmas > 1] -= 1
        print(mid_points)
        print(sigmas)

        resMask = np.zeros(resMap.shape)
        # sigmas = [1, 2, 3]
        for idx in range(0, edges.size - 1):
            sigma = sigmas[idx]
            mask = ((resMap >= edges[idx]) * (resMap < edges[idx + 1])).astype(int)
            aux = np.zeros(resMap.shape)
            aux[::sigma, ::sigma, ::sigma] = mask[::sigma, ::sigma, ::sigma]
            mask = sigma * aux
            # mask *= sigma
            resMask += mask
        resMask *= oriMask

        ImageHandler().write(resMask, self._getTmpPath("resMask.mrc"), overwrite=True)

        self.sigmas = ' '.join(map(str, np.unique(sigmas)))
        print(self.sigmas)

    #--------------------------- STEPS functions --------------------------------------------
    def convertInputStep(self):
        particlesMd = self._getTmpPath('corrected_particles.xmd')
        imgSet = self.inputParticles.get()

        def zernikeRow(part, partRow, **kwargs):
            imageToRow(part, partRow, emlib.MDL_IMAGE, **kwargs)
            coord = part.getCoordinate()
            if coord is not None:
                coordinateToRow(coord, partRow, copyId=False)
            if part.hasMicId():
                partRow.setValue(emlib.MDL_MICROGRAPH_ID, int(part.getMicId()))
                partRow.setValue(emlib.MDL_MICROGRAPH, str(part.getMicId()))
            if isinstance(part, ParticleFlex) and self.useZernike.get():
                z_clnm = part.getZFlex()
                partRow.setValue(emlib.MDL_SPH_COEFFICIENTS, z_clnm.tolist())

        writeSetOfImages(imgSet, particlesMd, zernikeRow)

        # Correct CTF of particles if needed
        if not self.ctfCorrected.get():
            sr = imgSet.getSamplingRate()
            corrected_stk = self._getTmpPath('corrected_particles.stk')
            args = "-i %s -o %s --save_metadata_stack --keep_input_columns --sampling_rate %f --correct_envelope" \
                   % (particlesMd, corrected_stk, sr)
            program = 'xmipp_ctf_correct_wiener2d'
            self.runJob(program, args, numberOfMpi=self.numberOfThreads.get(), env=xmipp3.Plugin.getEnviron())

        # Mask preprocessing (if provided)
        # if self.mask.get():
        #     mask_file = self.mask.get().getFileName()
        #     mask_zart_file = self._getTmpPath('mask_zart.vol')
        #     args = "--input %s --output %s" \
        #            % (mask_file, mask_zart_file)
        #     program = os.path.join(const.XMIPP_SCRIPTS, "flood_fill_mask.py")
        #     program = flexutils.Plugin.getProgram(program)
        #     self.runJob(program, args, numberOfMpi=1)

        # Prepare data (reconstruction mode)
        if self.mode.get() != 0:
            halvesMd = self._getTmpPath("corrected_particles_half_")
            self.runJob("xmipp_metadata_split",
                        "-i %s --oroot %s" % (particlesMd, halvesMd), env=xmipp3.Plugin.getEnviron())

    def createOutputStep(self):
        imgSet = self.inputParticles.get()
        volume = Volume()
        volume.setFileName(self._getExtraPath("final_reconstruction.mrc"))
        volume.setSamplingRate(imgSet.getSamplingRate())

        if self.mode.get() != 0:
            halves = self.volumeRestoration()
            volume.setHalfMaps(halves)
        
        self._defineOutputs(outputVolume=volume)
        self._defineSourceRelation(self.inputParticles, volume)
    
    #--------------------------- INFO functions -------------------------------------------- 
    def _summary(self):
        """ Should be overriden in subclasses to 
        return summary message for NORMAL EXECUTION. 
        """
        return []
    
    #--------------------------- UTILS functions --------------------------------------------
    def defineZARTArgs(self, inputMd, outFile, niter, step, mask):
        useGPU = self.useGpu.get()
        params = ' -i %s' % inputMd
        params += ' -o %s' % outFile
        params += ' --odir %s' % self._getExtraPath()
        params += ' --step %d' % step
        if "_level_" in outFile:
            reg = self.reg.get()
            level = float(re.findall(r'\d+', outFile)[1]) - 2
            if level > 0:
                params += " --regularization %f" % (reg * 0.9**level)
            else:
                params += " --regularization %f" % (reg)
        else:
            params += " --regularization %f" % self.reg.get()
        if hasattr(self, 'sigmas'):
            params += ' --sigma "' + self.sigmas + '"'
        else:
            params += ' --sigma "1.5"'
        params += ' --niter %d' % niter

        if isinstance(self.inputParticles.get(), SetOfParticlesFlex) and self.useZernike.get():
            particles = self.inputParticles.get()
            params += " --useZernike --l1 %d --l2 %d" % (particles.getFlexInfo().L1.get(),
                                                         particles.getFlexInfo().L2.get())
            # zernikeMask = particles.recMask.get() if hasattr(particles, "recMask") else self.mask.get()
            # if zernikeMask:
            #     params += " --mask %s" % zernikeMask

        if mask:
            params += ' --maskf %s --maskb %s' % (mask, mask)

        # GPU parameters
        if useGPU:
            onlyPositive = self.onlyPositive.get()
            params += ' --dThr %f' % self.dThr.get()
            if onlyPositive:
                params += " --onlyPositive"

        return params

    def volumeRestoration(self):
        half_1 = glob.glob(self._getExtraPath("final_reconstruction_1*"))
        half_2 = glob.glob(self._getExtraPath("final_reconstruction_2*"))
        half_1 = max(half_1, key=os.path.getctime)
        half_2 = max(half_2, key=os.path.getctime)

        params = "--i1 %s --i2 %s --oroot %s" % (half_1, half_2, self._getExtraPath("volume"))
        self.runJob('xmipp_volume_halves_restoration', params, numberOfMpi=1, env=xmipp3.Plugin.getEnviron())

        ih = ImageHandler()
        ih.convert(self._getExtraPath("volume_restored1.vol"), self._getExtraPath("final_reconstruction.mrc"),
                   overwrite=True)

        return [half_1, half_2]



# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *              James Krieger (jmkrieger@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import numpy as np
import os
from xmipp_metadata.image_handler import ImageHandler

from pwem.protocols import ProtAnalysis3D, ProtFlexBase
from pwem.objects import VolumeFlex, AtomStructFlex

import pyworkflow.protocol.params as params
import pyworkflow.utils as pwutils
from pyworkflow.object import Float

import xmipp3

import flexutils
import flexutils.constants as const
from flexutils.protocols.xmipp.utils.utils import computeBasis, readMap, getCoordsAtLevel, \
    getXmippOrigin, resizeZernikeCoefficients


class XmippApplyFieldZernike3D(ProtAnalysis3D, ProtFlexBase):
    """ Protocol for PDB deformation based on Zernike3D basis. """
    _label = 'apply deformation field - Zernike3D'

    # --------------------------- DEFINE param functions -----------------------
    def _defineParams(self, form):
        form.addSection(label='Input')
        form.addParam('inputVolume', params.PointerParam, label="Zernike3D volume(s)",
                      important=True, pointerClass="SetOfVolumesFlex, VolumeFlex",
                      help='Volume(s) with Zernike3D coefficients assigned.')
        form.addParam('applyPDB', params.BooleanParam, label="Apply to structure?",
                      default=False,
                      help="If True, you will be able to provide an atomic structure to be deformed "
                           "based on the Zernike3D coefficients associated to the input volume(s). "
                           "If False, the coefficients will be applied to the volume(s) directly.")
        form.addParam('inputPDB', params.PointerParam, label="Input PDB",
                      pointerClass='AtomStruct', allowsNull=True, condition="applyPDB==True",
                      help='Atomic structure to apply the deformation fields defined by the '
                           'Zernike3D coefficients associated to the input volume. '
                           'For better results, the volume(s) and structure should be aligned')
        form.addParam('moveBoxOrigin', params.BooleanParam, default=False, condition="applyPDB==True",
                      label="Move structure to box origin?",
                      help="If PDB has been aligned inside Scipion, set to False. Otherwise, this option will "
                           "correctly place the PDB in the origin of the volume.")

    # --------------------------- INSERT steps functions -----------------------
    def _insertAllSteps(self):
        self._insertFunctionStep(self.deformStep)
        self._insertFunctionStep(self.analyzeRSStep)
        self._insertFunctionStep(self.createOutputStep)

    # --------------------------- STEPS functions ------------------------------
    def deformStep(self):
        volumes = self.inputVolume.get()
        if isinstance(volumes, VolumeFlex):
            volumes = [volumes]
            num_vols = 1
        else:
            volumes = volumes
            num_vols = volumes.getSize()

        self.len_num_vols = len(str(num_vols))

        idx = 0
        for volume in volumes:
            i_pad = str(idx).zfill(self.len_num_vols)

            # Write coefficients to file
            z_clnm_file = self._getExtraPath("z_clnm_{0}.txt".format(i_pad))
            z_clnm = volume.getZFlex()

            # if self.applyPDB.get():
            #     z_clnm *= volume.getSamplingRate()

            z_clnm = np.char.mod('%f', z_clnm)
            z_clnm = ",".join(z_clnm)

            self.writeZernikeFile(volume, z_clnm, z_clnm_file)

            samplingRate = volume.getSamplingRate()
            boxsize = volume.getXDim()
            vol_file = volume.getFlexInfo().refMap.get()
            if self.applyPDB.get():
                ref_file = self.inputPDB.get().getFileName()
                outFile = pwutils.removeBaseExt(self.inputPDB.get().getFileName()) + '_deformed_{0}.pdb'.format(i_pad)
                outFile = self._getExtraPath(outFile)
            else:
                ref_file = volume.getFlexInfo().refMask.get()
                outFile = self._getExtraPath("deformed_volume_{0}.mrc".format(i_pad))

            args = "--ref_file %s --vol_file %s --z_file %s --out_file %s --boxsize %d --sr %f" \
                   % (ref_file, vol_file, z_clnm_file, outFile, boxsize, samplingRate)
            program = "apply_deformation_field_zernike3d.py"
            program = flexutils.Plugin.getProgram(program)
            self.runJob(program, args, env=xmipp3.Plugin.getEnviron())

            idx += 1

    def analyzeRSStep(self):
        volumes = self.inputVolume.get()

        if isinstance(volumes, VolumeFlex):
            volumes = [volumes]
            num_vols = 1
            first_vol = volumes[0]
        else:
            volumes = volumes
            num_vols = volumes.getSize()
            first_vol = volumes.getFirstItem()

        self.len_num_vols = len(str(num_vols))

        # Get reference coords and indices
        ref_file = first_vol.getFlexInfo().refMask.get()
        ref_map = readMap(ref_file)
        coords = getCoordsAtLevel(ref_map, 1)
        origin = getXmippOrigin(ref_map)
        coords_xo = coords - origin
        indices = np.transpose(np.asarray([coords[:, 2], coords[:, 1], coords[:, 0]]))

        # Get Zernike3D basis
        Rmax = first_vol.getFlexInfo().Rmax.get()
        L1 = first_vol.getFlexInfo().L1.get()
        L2 = first_vol.getFlexInfo().L2.get()
        Z = computeBasis(L1=L1, L2=L2, pos=coords_xo, r=Rmax)

        # Save indices
        indices_path = self._getExtraPath("indices.txt")
        np.savetxt(indices_path, indices)

        idx = 0
        for volume in volumes:
            i_pad = str(idx).zfill(self.len_num_vols)

            # Get z_clnm
            z_clnm = volume.getZFlex()
            A = resizeZernikeCoefficients(z_clnm)

            # Get deformation field
            d_f = Z @ A.T

            # Save deformation field
            field_path = self._getExtraPath("def_field.txt")
            np.savetxt(field_path, d_f)

            # RS analysis
            args = "--field %s --indices %s --out_path %s --boxsize %d" % \
                   (field_path, indices_path, self._getExtraPath(), ref_map.shape[0])
            program = "strain_rotation_analysis.py"
            program = flexutils.Plugin.getProgram(program)
            self.runJob(program, args)

            if self.applyPDB.get():
                struct_file = self.inputPDB.get().getFileName()
                program = 'xmipp_pdb_label_from_volume'

                # Strain labeling
                outFile = self._getExtraPath(
                    pwutils.removeBaseExt(struct_file) + '_deformed_{0}_strain.pdb'.format(i_pad)
                )
                args = "--pdb %s --vol %s -o %s --sampling %f --radius 5 --origin %d %d %d" % \
                       (struct_file, self._getExtraPath("strain.mrc"), outFile, volume.getSamplingRate(),
                        origin[0], origin[0], origin[0])
                self.runJob(program, args, env=xmipp3.Plugin.getEnviron())

                # Rotation labeling
                outFile = self._getExtraPath(
                    pwutils.removeBaseExt(struct_file) + '_deformed_{0}_rotation.pdb'.format(i_pad)
                )
                args = "--pdb %s --vol %s -o %s --sampling %f --radius 5 --origin %d %d %d" % \
                       (struct_file, self._getExtraPath("rotation.mrc"), outFile, volume.getSamplingRate(),
                        origin[0], origin[0], origin[0])
                self.runJob(program, args, env=xmipp3.Plugin.getEnviron())

            else:
                # Rename strain and rotation volumes
                pwutils.moveFile(self._getExtraPath("strain.mrc"),
                                 self._getExtraPath("deformed_volume_{0}_strain.mrc".format(i_pad)))
                pwutils.moveFile(self._getExtraPath("rotation.mrc"),
                                 self._getExtraPath("deformed_volume_{0}_rotation.mrc".format(i_pad)))

            idx += 1

        # Cleaning to save some memory
        pwutils.cleanPattern(self._getExtraPath("*.txt"))

    def createOutputStep(self):
        volumes = self.inputVolume.get()
        Rmax = Float(int(0.5 * volumes.getXDim()))
        if isinstance(volumes, VolumeFlex):
            L1 = volumes.getFlexInfo().L1
            L2 = volumes.getFlexInfo().L2
            refMap = volumes.getFlexInfo().refMap
            refMask = volumes.getFlexInfo().refMask
            z_clnm = volumes.getZFlex()

            if self.applyPDB.get():
                outFile = pwutils.removeBaseExt(self.inputPDB.get().getFileName()) + '_deformed_0.pdb'
                pdb = AtomStructFlex(filename=self._getExtraPath(outFile), progName=const.ZERNIKE3D)
                pdb.getFlexInfo().L1 = L1
                pdb.getFlexInfo().L2 = L2
                pdb.getFlexInfo().Rmax = Float(volumes.getSamplingRate() * Rmax.get())
                pdb.getFlexInfo().refMap = refMap
                pdb.getFlexInfo().refMask = refMask
                pdb.setZFlex(z_clnm)
                self._defineOutputs(deformed=pdb)
                self._defineSourceRelation(self.inputPDB, pdb)
                self._defineSourceRelation(self.inputVolume, pdb)
            else:
                vol = VolumeFlex(progName=const.ZERNIKE3D)
                vol.setSamplingRate(volumes.getSamplingRate())
                vol.setFileName(self._getExtraPath("deformed_volume_0.mrc"))
                vol.getFlexInfo().L1 = L1
                vol.getFlexInfo().L2 = L2
                vol.getFlexInfo().Rmax = Rmax
                vol.getFlexInfo().refMap = refMap
                vol.getFlexInfo().refMask = refMask
                vol.setZFlex(z_clnm)
                self._defineOutputs(deformed=vol)
                self._defineSourceRelation(volumes, vol)
        else:
            L1 = volumes.getFirstItem().getFlexInfo().L1
            L2 = volumes.getFirstItem().getFlexInfo().L2
            refMap = volumes.getFirstItem().getFlexInfo().refMap
            refMask = volumes.getFirstItem().getFlexInfo().refMask
            if self.applyPDB.get():
                pdbs = self._createSetOfAtomStructFlex(progName=const.ZERNIKE3D)
            else:
                vols = self._createSetOfVolumesFlex(progName=const.ZERNIKE3D)
                vols.setSamplingRate(volumes.getSamplingRate())

            idx = 0
            for volume in volumes:
                i_pad = str(idx).zfill(self.len_num_vols)

                z_clnm = volume.getZFlex()

                if self.applyPDB.get():
                    outFile = pwutils.removeBaseExt(self.inputPDB.get().getFileName()) + '_deformed_{0}.pdb'.format(i_pad)
                    pdb = AtomStructFlex(filename=self._getExtraPath(outFile), progName=const.ZERNIKE3D)
                    pdb.getFlexInfo().L1 = L1
                    pdb.getFlexInfo().L2 = L2
                    pdb.getFlexInfo().Rmax = Float(volume.getSamplingRate() * Rmax.get())
                    pdb.getFlexInfo().refMap = refMap
                    pdb.getFlexInfo().refMask = refMask
                    pdb.setZFlex(z_clnm)

                    pdbs.append(pdb)
                else:
                    vol = VolumeFlex(progName=const.ZERNIKE3D)
                    vol.setSamplingRate(volume.getSamplingRate())
                    vol.setFileName(self._getExtraPath("deformed_volume_{0}.mrc".format(i_pad)))
                    vol.getFlexInfo().L1 = L1
                    vol.getFlexInfo().L2 = L2
                    vol.getFlexInfo().Rmax = Rmax
                    vol.getFlexInfo().refMap = refMap
                    vol.getFlexInfo().refMask = refMask
                    vol.setZFlex(z_clnm)

                    vols.append(vol)

                idx += 1

            if self.applyPDB.get():
                self._defineOutputs(deformed=pdbs)
                self._defineSourceRelation(self.inputPDB, pdbs)
                self._defineSourceRelation(self.inputVolume, pdbs)
            else:
                self._defineOutputs(deformed=vols)
                self._defineSourceRelation(self.inputVolume, vols)

    # --------------------------- UTILS functions ------------------------------
    def writeZernikeFile(self, volume, z_clnm, file):
        L1 = volume.getFlexInfo().L1.get()
        L2 = volume.getFlexInfo().L2.get()
        Rmax = int(0.5 * volume.getXDim())
        # Rmax = volume.getSamplingRate() * Rmax if self.applyPDB.get() else Rmax
        with open(file, 'w') as fid:
            fid.write(' '.join(map(str, [L1, L2, Rmax])) + "\n")
            fid.write(z_clnm.replace(",", " ") + "\n")



# **************************************************************************
# *
# * Authors:     David Herreros (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
from pyworkflow.utils import weakImport

from .protocol_reconstruct_zart import XmippProtReconstructZART
from .protocol_match_and_deform_structure_zernike3d import XmippMatchDeformSructZernike3D
from .protocol_match_and_deform_map_zernike3d import XmippMatchDeformMapZernike3D
from .protocol_assign_heterogeneity_priors_zernike3d import XmippProtHeterogeneityPriorsZernike3D
from .protocol_angular_alignment_zernike3d import XmippProtAngularAlignmentZernike3D
from .protocol_focus_zernike3d import XmippProtFocusZernike3D
from .protocol_reassign_reference_zernike3d import XmippProtReassignReferenceZernike3D
# from .protocol_compute_priors_zernike3d import XmippProtComputeHeterogeneityPriorsZernike3D
from .protocol_statistics_zernike3d import XmippProtStatisticsZernike3D
from .protocol_structure_landscape import XmippProtStructureLanscapes
from .protocol_cluster_structures_zernike3d import XmippProtClusterStructuresZernike3D
from .protocol_apply_field_zernike3d import XmippApplyFieldZernike3D
# from .protocol_apply_field_nma import XmippApplyFieldNMA
from .protocol_prepare_volumes_zernike3deep import XmippProtPrepareZernikeVolumes

# Weak imports to avoid library incompatibilities in viewers due to Xmipp environment (this avoids loading
# the Xmipp env and the Scipion libraries that interfere with flexutils ones)
with weakImport("xmipp3.base"):
    from .protocol_resize_zernike_data import XmippProtCropResizeZernikeParticles, XmippProtCropResizeZernikeVolumes

from .protocol_deform_map_zernike3d import ProtFlexVolumeDeformZernike3D
# from .protocol_structure_map_zernike3d import XmippProtStructureMapZernike3D

from .protocol_angular_align_zernike3deep import TensorflowProtAngularAlignmentZernike3Deep
from .protocol_predict_zernike3deep import TensorflowProtPredictZernike3Deep
from .protocol_angular_align_deep_pose import TensorflowProtAngularAlignmentDeepPose
from .protocol_predict_deep_pose import TensorflowProtPredictDeepPose
# from .protocol_angular_align_deep_nma import TensorflowProtAngularAlignmentDeepNMA
# from .protocol_predict_deep_nma import TensorflowProtPredictDeepNMA
from .protocol_angular_align_homo_siren import TensorflowProtAngularAlignmentHomoSiren
from .protocol_predict_homo_siren import TensorflowProtPredictHomoSiren
from .protocol_angular_align_het_siren import TensorflowProtAngularAlignmentHetSiren
from .protocol_predict_het_siren import TensorflowProtPredictHetSiren
from .protocol_denoise_particles_het_siren import TensorflowProtDenoiseParticlesHetSiren
from .protocol_train_flex_consensus import TensorflowProtTrainFlexConsensus
from .protocol_interactive_flex_consensus import TensorflowProtInteractiveFlexConsensus



# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import numpy as np
import prody as pd
from pathlib import Path
import os

from pwem.protocols import ProtAnalysis3D, ProtFlexBase
from pwem.objects import AtomStructFlex

import pyworkflow.protocol.params as params
import pyworkflow.utils as pwutils

import xmipp3

import flexutils
import flexutils.constants as const
from flexutils.protocols.xmipp.utils.utils import inscribedRadius

class XmippApplyFieldNMA(ProtAnalysis3D, ProtFlexBase):
    """ Protocol for PDB deformation based on NMA basis. """
    _label = 'apply deformation field - NMA'
    _subset = ["ca", "bb", None]

    # --------------------------- DEFINE param functions -----------------------
    def _defineParams(self, form):
        form.addSection(label='Input')
        form.addParam('inputStruct', params.PointerParam, label="NMA structure(s)",
                      important=True, pointerClass="SetOfAtomStructFlex, AtomStructFlex",
                      help='Structure(s) with NMA coefficients assigned.')

    # --------------------------- INSERT steps functions -----------------------
    def _insertAllSteps(self):
        self._insertFunctionStep(self.deformStep)
        self._insertFunctionStep(self.analyzeRSStep)
        self._insertFunctionStep(self.createOutputStep)

    # --------------------------- STEPS functions ------------------------------
    def deformStep(self):
        structs = self.inputStruct.get()
        if isinstance(structs, AtomStructFlex):
            structs = [structs]
            num_structs = 1
        else:
            structs = structs
            num_structs = structs.getSize()
        base_name = pwutils.removeBaseExt(structs.getFirstItem().getFileName())

        self.len_num_structs = len(str(num_structs))

        # Load NMA basis
        basis_file = Path(Path(structs.getFirstItem().getFlexInfo().modelPath.get()).parent.parent, "nma_basis.anm.npz")
        anm = pd.loadModel(str(basis_file))
        U = anm.getEigvecs()

        # Read structure file and get coords
        subset = structs.getFirstItem().getFlexInfo().atomSubset.get()
        pd_struct = pd.parsePDB(structs.getFirstItem().getFileName(), subset=subset, compressed=False)
        coords = pd_struct.getCoords()

        # Get and save boxsize (for strain/rotation analysis)
        boxsize = np.ceil(2 * inscribedRadius(coords)).astype(int)
        boxsize_path = self._getExtraPath("boxsize.txt")
        with open(boxsize_path, "w") as fid:
            fid.write(str(boxsize))

        # Get and save indices (for strain/rotation analysis)
        coords_centered = coords - np.mean(coords, axis=0)
        indices = np.round(coords_centered + int(0.5 * boxsize)).astype(int)
        indices = np.transpose(np.vstack([indices[:, 2], indices[:, 1], indices[:, 0]]))  # Indices must be in ZYX order
        indices_path = self._getExtraPath("indices.txt")
        np.savetxt(indices_path, indices)

        idx = 0
        for struct in structs:
            i_pad = str(idx).zfill(self.len_num_structs)

            # Get deformation field
            c_nma = struct.getZFlex()
            d_f = (U @ c_nma.T).reshape(-1, 3)

            # Apply deformation field
            c_moved = coords + d_f

            # Save deformation field (for strain/rotation analysis)
            field_path = self._getExtraPath("def_field_{0}.txt".format(i_pad))
            np.savetxt(field_path, d_f)

            # Saved deformed structure
            out_struct = pd_struct.copy()
            out_struct.setCoords(c_moved)
            pd.writePDB(self._getExtraPath(base_name + '_deformed_{0}.pdb'.format(i_pad)), out_struct)

            idx += 1

    def analyzeRSStep(self):
        indices_path = self._getExtraPath("indices.txt")
        with open(self._getExtraPath("boxsize.txt"), "r") as fid:
            boxsize = int(fid.readlines()[0])
        origin = np.floor(0.5 * boxsize).astype(int)

        structs = self.inputStruct.get()
        if isinstance(structs, AtomStructFlex):
            num_structs = 1
        else:
            num_structs = structs.getSize()

        struct_file = structs.getFirstItem().getFileName()
        base_name = pwutils.removeBaseExt(struct_file)

        for idx in range(num_structs):
            i_pad = str(idx).zfill(self.len_num_structs)

            # Get paths for this structure
            field_path = self._getExtraPath("def_field_{0}.txt".format(i_pad))
            out_path_strain = self._getExtraPath(base_name + '_deformed_{0}_strain.pdb'.format(i_pad))
            out_path_rotation = self._getExtraPath(base_name + '_deformed_{0}_rotation.pdb'.format(i_pad))

            # RS analysis
            args = "--field %s --indices %s --out_path %s --boxsize %d" % \
                   (field_path, indices_path, self._getExtraPath(), boxsize)
            program = "strain_rotation_analysis.py"
            program = flexutils.Plugin.getProgram(program)
            self.runJob(program, args)

            # Strain labeling
            program = 'xmipp_pdb_label_from_volume'
            args = "--pdb %s --vol %s -o %s --sampling 1.0 --radius 5 --origin %d %d %d" % \
                   (struct_file, self._getExtraPath("strain.mrc"), out_path_strain,
                    origin, origin, origin)
            self.runJob(program, args, env=xmipp3.Plugin.getEnviron())

            # Rotation labeling
            args = "--pdb %s --vol %s -o %s --sampling 1.0 --radius 5 --origin %d %d %d" % \
                   (struct_file, self._getExtraPath("rotation.mrc"), out_path_rotation,
                    origin, origin, origin)
            self.runJob(program, args, env=xmipp3.Plugin.getEnviron())

        # Cleaning to save some memory
        pwutils.cleanPattern(self._getExtraPath("*.txt"))

    def createOutputStep(self):
        structs = self.inputStruct.get()
        base_name = pwutils.removeBaseExt(structs.getFirstItem().getFileName())

        if isinstance(structs, AtomStructFlex):
            c_nma = structs.getZFlex()

            outFile = self._getExtraPath(base_name + '_deformed.pdb')
            out_struct = AtomStructFlex(progName=const.NMA)
            out_struct.copyInfo(structs)
            out_struct.setLocation(outFile)
            out_struct.setZFlex(c_nma)

            self._defineOutputs(deformed=out_struct)
            self._defineSourceRelation(self.inputStruct, out_struct)
        else:
            out_structs = self._createSetOfAtomStructFlex(progName=const.NMA)

            idx = 0
            for struct in structs:
                i_pad = str(idx).zfill(self.len_num_structs)

                c_nma = struct.getZFlex()

                outFile = self._getExtraPath(base_name + '_deformed_{0}.pdb'.format(i_pad))
                out_struct = AtomStructFlex(progName=const.NMA)
                out_struct.copyInfo(struct)
                out_struct.setLocation(outFile)
                out_struct.setZFlex(c_nma)

                out_structs.append(out_struct)

                idx += 1

            self._defineOutputs(deformed=out_structs)
            self._defineSourceRelation(self.inputStruct, out_structs)

    # --------------------------- UTILS functions ------------------------------



# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import numpy as np
import os
import pickle
from xmipp_metadata.image_handler import ImageHandler

from sklearn.neighbors import KDTree
from sklearn.cluster import KMeans

import pyvista as pv

from traits.api import HasTraits, Instance, Array, Float, Bool, String,\
    on_trait_change, Callable
from traits.trait_types import Button, Enum
from traitsui.api import View, Item, HGroup, Group, HSplit, VGroup, RangeEditor, TextEditor, EnumEditor

from pyface.image_resource import ImageResource

from tvtk.api import tvtk
from tvtk.pyface.scene import Scene

from mayavi import mlab
from mayavi.core.api import PipelineBase, Source
from mayavi.core.ui.api import SceneEditor, MayaviScene, \
    MlabSceneModel

import flexutils
from flexutils.viewers.chimera_viewers.viewer_morph_chimerax import FlexMorphChimeraX


################################################################################
# The object implementing the dialog
class VolumeSlicer(HasTraits):
    # The data to plot
    data = Array()
    grid = Array()

    # The values to interpolate
    interp_val = Array()

    # The 4 views displayed (Zernike space)
    scene3d = Instance(MlabSceneModel, ())
    scene_x = Instance(MlabSceneModel, ())
    scene_y = Instance(MlabSceneModel, ())
    scene_z = Instance(MlabSceneModel, ())

    # Real time conformation view
    scene_c = Instance(MlabSceneModel, ())

    # The data source
    data_src3d = Instance(Source)

    # The image plane widgets of the 3D scene
    ipw_3d_x = Instance(PipelineBase)
    ipw_3d_y = Instance(PipelineBase)
    ipw_3d_z = Instance(PipelineBase)

    # Line cuts for 3d view
    line_3d_x = Instance(PipelineBase)
    line_3d_y = Instance(PipelineBase)
    line_3d_z = Instance(PipelineBase)

    # Point cloud (actor)
    ipw_pc = Instance(PipelineBase)
    ipw_sel = Instance(PipelineBase)

    # Map (actor)
    ipw_map = Instance(PipelineBase)

    # Visualization style
    show_cut_planes = Bool(True)
    opacity = Float(0.0)
    threshold = Float(0.0)
    op_min = Float(0.0)
    op_max = Float(1.0)

    _axis_names = dict(x=0, y=1, z=2)

    # Output path
    path = String()

    # Zernike parameters
    z_space = Array()
    vector_file = String('z_vw.txt')

    # For picking coefficients
    cursor_position = Array()

    # Map
    map = Array()
    generated_map = Array()

    # Mask
    mask_file = String()

    # KMeans
    n_clusters = String("10")

    # Selections
    selection_name = String()
    # show_selections = Bool(True)
    current_sel = String('origin')
    ipw_label = Instance(PipelineBase)
    save_file = String('saved_selections.txt')

    # Save selections
    add_selection = Button("Add conformation")
    save_selections = Button("Save selections")
    remove_selection = Button("Remove conformation")
    rename_selection = Button("Rename selection")
    compute_kmeans = Button("Compute KMeans")

    # Generate map function
    mode = String()
    generate_map = Callable()

    # Points properties
    scale_pts = Button("Scale spheres")
    scale_factor = String("3")

    # ChimeraX morphing
    morphing_choice = Enum('Salesman', 'Random walk')
    morph_chimerax = Button(label="",
                            image=ImageResource(os.path.join(os.path.dirname(flexutils.__file__), "chimerax_logo.png")))

    # ---------------------------------------------------------------------------
    def __init__(self, **traits):
        self.class_inputs = traits
        super(VolumeSlicer, self).__init__(**traits)
        # Force the creation of the image_plane_widgets:
        self.interpolateGrid()
        self.moveDataToGrid()
        self.map
        self.ipw_3d_x
        self.ipw_3d_y
        self.ipw_3d_z
        self.ipw_pc
        self.ipw_map
        self.line_3d_x
        self.line_3d_y
        self.line_3d_z
        self.path

        # Create KDTree
        self.kdtree_data = KDTree(self.data)
        self.kdtree_z_space = KDTree(self.z_space)

        # Generate map
        self.mode
        if self.mode == "Zernike3D":
            from flexutils.protocols.xmipp.utils.utils import applyDeformationField
            self.generate_map = applyDeformationField
        elif self.mode == "CryoDrgn":
            import cryodrgn
            from cryodrgn.utils import generateVolumes
            cryodrgn.Plugin._defineVariables()
            self.generate_map = generateVolumes
        elif self.mode == "HetSIREN":
            from flexutils.utils import generateVolumesHetSIREN
            self.generate_map = generateVolumesHetSIREN
        elif self.mode == "NMA":
            from flexutils.utils import generateVolumesDeepNMA
            self.generate_map = generateVolumesDeepNMA

        # Selections
        self.selections = dict()
        pathFile = os.path.join(self.path, "selections_dict.pkl")
        if os.path.isfile(pathFile):
            self.readSelectionsDict()

        _, idx = self.kdtree_data.query(np.asarray([0, 0, 0]).reshape(1, -1), k=1)
        self.selections['origin'] = idx[0][0]
        if not os.path.isfile(pathFile) and self.mode == "Zernike3D":
            for idx in range(int(self.class_inputs["n_vol"])):
                self.selections['class_%d' % (idx + 1)] = self.data.shape[0] - (idx + 1)

        self.ipw_sel

    # ---------------------------------------------------------------------------
    # Default values
    # ---------------------------------------------------------------------------
    def _data_src3d_default(self):
        return mlab.pipeline.scalar_field(self.grid, figure=self.scene3d.mayavi_scene, colormap="viridis")

    def make_ipw_3d(self, axis_name):
        ipw = mlab.pipeline.image_plane_widget(self.data_src3d,
                                               figure=self.scene3d.mayavi_scene,
                                               transparent=True,
                                               colormap="black-white",
                                               plane_orientation='%s_axes' % axis_name)
        return ipw

    def make_line_3d(self, axis_name):
        if axis_name == "x":
            normal = np.array([1, 0, 0])
        elif axis_name == "y":
            normal = np.array([0, 1, 0])
        elif axis_name == "z":
            normal = np.array([0, 0, 1])
        white = (1, 1, 1)
        grid_size = np.asarray(self.grid.shape)
        point1 = 0.5 * grid_size * (1 - normal)
        point2 = grid_size * normal + point1
        x = [point1[0], point2[0]]
        y = [point1[1], point2[1]]
        z = [point1[2], point2[2]]
        line = mlab.plot3d(x, y, z, color=white, tube_radius=1., figure=self.scene3d.mayavi_scene)
        return line

    def _ipw_3d_x_default(self):
        return self.make_ipw_3d('x')

    def _ipw_3d_y_default(self):
        return self.make_ipw_3d('y')

    def _ipw_3d_z_default(self):
        return self.make_ipw_3d('z')

    def _line_3d_x_default(self):
        return self.make_line_3d('x')

    def _line_3d_y_default(self):
        return self.make_line_3d('y')

    def _line_3d_z_default(self):
        return self.make_line_3d('z')

    # def _ipw_pc_default(self):
    #     return self.display_space_cloud()
    #
    # def _ipw_sel_default(self):
    #     return self.display_selections()
    #
    # def _ipw_map_default(self):
    #     return self.display_map()

    def _map_default(self):
        # map = self.readMap(self.map_file)
        return np.zeros([64, 64, 64])

    # ---------------------------------------------------------------------------
    # Scene activation callbaks
    # ---------------------------------------------------------------------------
    @on_trait_change('scene3d.activated')
    def display_scene3d(self):
        # outline = mlab.pipeline.outline(self.data_src3d,
        #                                 figure=self.scene3d.mayavi_scene,
        #                                 )
        self.scene3d.mlab.view(40, 50)
        # Interaction properties can only be changed after the scene
        # has been created, and thus the interactor exists
        for ipw in (self.ipw_3d_x, self.ipw_3d_y, self.ipw_3d_z):
            # Turn the interaction off
            ipw.ipw.interaction = 0
        self.scene3d.scene.background = (0, 0, 0)
        # Keep the view always pointing up
        self.scene3d.scene.interactor.interactor_style = \
            tvtk.InteractorStyleTerrain()

        # Pick event
        self.scene3d.scene.interactor.add_observer('KeyPressEvent', self.on_pick)
        self.picker = self.scene3d.mayavi_scene.on_mouse_pick(self.picker_callback)
        # Decrease the tolerance, so that we can more easily select a precise
        # point.
        self.picker.tolerance = 0.01

    @on_trait_change('scene_c.activated')
    def display_scene_c(self):
        self.scene_c.mlab.view(40, 50)
        self.scene_c.scene.background = (0, 0, 0)

    @on_trait_change('scene3d.activated')
    def display_space_cloud(self):
        scatter = mlab.pipeline.scalar_scatter(self.data[:, 2], self.data[:, 1], self.data[:, 0], self.interp_val,
                                               figure=self.scene3d.mayavi_scene)
        scatter = mlab.pipeline.glyph(scatter,
                                      opacity=self.opacity,
                                      scale_mode='none', scale_factor=3, mode='sphere', colormap="viridis",
                                      figure=self.scene3d.mayavi_scene)
        scatter.actor.actor.pickable = 0
        setattr(self, 'ipw_pc', scatter)

    @on_trait_change('scene3d.activated')
    def display_selections(self):
        data = self.data[list(self.selections.values())]
        scatter = mlab.points3d(data[:, 2], data[:, 1], data[:, 0],
                                scale_mode='none', scale_factor=3, mode='sphere', color=(1, 1, 1),
                                figure=self.scene3d.mayavi_scene)
        setattr(self, 'ipw_sel', scatter)

    @on_trait_change('scene3d.activated')
    def populateLabels(self):
        self.ipw_label = self.addLabel('origin')

    def addLabel(self, label):
        data = self.data[self.selections[label]]
        text = mlab.text3d(data[2] + 5., data[1] + 5., data[0] + 5., label, scale=3, color=(1, 1, 1),
                           figure=self.scene3d.mayavi_scene)
        return text

    @on_trait_change('scene_c.activated')
    def display_map(self):
        # Possible visualization
        # volume = mlab.pipeline.scalar_field(self.map, figure=self.scene_c.mayavi_scene)
        # volume = mlab.pipeline.volume(volume, figure=self.scene_c.mayavi_scene, color=(1, 1, 1),
        #                               vmin=0.0, vmax=0.05)

        # Chimera visualization (slow)
        volume = mlab.contour3d(self.map, color=(1, 1, 1), contours=1, figure=self.scene_c.mayavi_scene)
        volume.contour.auto_contours = False
        volume.contour.auto_update_range = True
        setattr(self, 'ipw_map', volume)

    def make_side_view(self, axis_name):
        scene = getattr(self, 'scene_%s' % axis_name)

        # To avoid copying the data, we take a reference to the
        # raw VTK dataset, and pass it on to mlab. Mlab will create
        # a Mayavi source from the VTK without copying it.
        # We have to specify the figure so that the data gets
        # added on the figure we are interested in.
        outline = mlab.pipeline.outline(
            self.data_src3d.mlab_source.dataset,
            figure=scene.mayavi_scene,
        )
        ipw = mlab.pipeline.image_plane_widget(
            outline,
            colormap="viridis",
            plane_orientation='%s_axes' % axis_name)
        setattr(self, 'ipw_%s' % axis_name, ipw)

        # Synchronize positions between the corresponding image plane
        # widgets on different views.
        ipw.ipw.sync_trait('slice_position',
                           getattr(self, 'ipw_3d_%s' % axis_name).ipw)

        # Make left-clicking create a crosshair
        ipw.ipw.left_button_action = 0

        # Add a callback on the image plane widget interaction to
        # move the others
        def move_view(obj, evt):
            # Update 3D plane position
            position = obj.GetCurrentCursorPosition()
            self.moveViewToPosition(position, axis_name)



        ipw.ipw.add_observer('InteractionEvent', move_view)
        ipw.ipw.add_observer('StartInteractionEvent', move_view)

        # Center the image plane widget
        ipw.ipw.slice_position = 0.5 * self.grid.shape[
            self._axis_names[axis_name]]

        # Position the view for the scene
        views = dict(x=(0, 90),
                     y=(90, 90),
                     z=(0, 0),
                     )
        scene.mlab.view(*views[axis_name])
        # 2D interaction: only pan and zoom
        scene.scene.interactor.interactor_style = \
            tvtk.InteractorStyleImage()
        scene.scene.background = (0, 0, 0)

    @on_trait_change('scene_x.activated')
    def display_scene_x(self):
        return self.make_side_view('x')

    @on_trait_change('scene_y.activated')
    def display_scene_y(self):
        return self.make_side_view('y')

    @on_trait_change('scene_z.activated')
    def display_scene_z(self):
        return self.make_side_view('z')

    @on_trait_change("opacity")
    def change_opacity_point_cloud(self):
        self.ipw_pc.actor.property.opacity = self.opacity

    @on_trait_change("threshold")
    def change_map_contour(self):
        volume = getattr(self, 'ipw_map')
        val_max = np.amax(volume.mlab_source.m_data.scalar_data)
        val_min = np.amin(volume.mlab_source.m_data.scalar_data)
        if val_max != val_min:
            contour = (val_max - val_min) * self.threshold + val_min
            volume.contour.contours = [contour]

    @on_trait_change("selection_name")
    def change_selection_label(self):
        self.ipw_label.trait_set(text=self.selection_name)
        self.ipw_label.actor.property.color = (0, 0.76, 1)

    @on_trait_change("n_clusters")
    def change_n_clusters(self):
        if not str(self.n_clusters).isdigit():
            self.n_clusters = "10"

    @on_trait_change("show_cut_planes")
    def update_cut_plane_visibility(self):
        self.ipw_3d_x.visible = self.show_cut_planes
        self.ipw_3d_y.visible = self.show_cut_planes
        self.ipw_3d_z.visible = self.show_cut_planes

    # @on_trait_change("show_selections")
    # def update_cut_plane_visibility(self):
    #     self.ipw_sel.visible = self.show_selections
    #     self.ipw_label.visible = self.show_selections

    def on_pick(self, obj, evt):
        if obj.GetKeyCode() == "c":
            pos = np.asarray([self.cursor_position[2],
                              self.cursor_position[1],
                              self.cursor_position[0]]).reshape(1, -1)
            _, ind = self.kdtree_data.query(pos, k=1)
            self.selections.append(self.z_space[ind, :])

    def _save_selections_fired(self):
        pathFile = os.path.join(self.path, self.save_file)
        with open(pathFile, 'w') as fid:
            for key, idc in self.selections.items():
                if "origin" not in key and "class" not in key:
                    vector = self.z_space[idc]
                    fid.write(' '.join(map(str, vector.reshape(-1))) + "\n")
        self.saveSelectionsDict()

    def _add_selection_fired(self):
        pos = np.asarray([self.cursor_position[2],
                          self.cursor_position[1],
                          self.cursor_position[0]]).reshape(1, -1)
        _, ind = self.kdtree_data.query(pos, k=1)
        key = "conf_%d" % (len(self.selections) + 1)
        self.selections[key] = ind[0, 0]

        data = self.data[list(self.selections.values())]
        self.ipw_sel.mlab_source.reset(x=data[:, 2], y=data[:, 1], z=data[:, 0])

    def _compute_kmeans_fired(self):
        # First delete all automatic selections associated to a previous KMeans
        for key in list(self.selections.keys()):
            if 'kmean' in key:
                del self.selections[key]
        # Compute KMeans and save automatic selection
        if int(self.n_clusters) > 0:
            clusters = KMeans(n_clusters=int(self.n_clusters)).fit(self.z_space)
            centers = clusters.cluster_centers_
            _, inds = self.kdtree_z_space.query(centers, k=1)
            for idx, ind in enumerate(inds):
                self.selections["kmean_%d" % (idx + 1)] = ind[0]
        data = self.data[list(self.selections.values())]
        self.ipw_sel.mlab_source.reset(x=data[:, 2], y=data[:, 1], z=data[:, 0])

    def _remove_selection_fired(self):
        del self.selections[self.current_sel]
        self.current_sel = ''
        data = self.data[list(self.selections.values())]
        self.ipw_sel.mlab_source.reset(x=data[:, 2], y=data[:, 1], z=data[:, 0])
        self.ipw_label.trait_set(text=self.current_sel)
        self.ipw_label.actor.property.color = (1, 1, 1)

    def _rename_selection_fired(self):
        if self.selection_name not in list(self.selections.keys()):
            self.selections[self.selection_name] = self.selections.pop(self.current_sel)
            self.current_sel = self.selection_name
            self.ipw_label.trait_set(text=self.current_sel)
            self.ipw_label.actor.property.color = (1, 1, 1)
            data = self.data[list(self.selections.values())]
            self.ipw_sel.mlab_source.reset(x=data[:, 2], y=data[:, 1], z=data[:, 0])

    def _scale_pts_fired(self):
        scale_factor = float(self.scale_factor)
        pts, pts_sel = getattr(self, 'ipw_pc'), getattr(self, 'ipw_sel')
        pts.glyph.glyph.scale_factor = scale_factor
        pts_sel.glyph.glyph.scale_factor = scale_factor + 1.
        self.ipw_label.scale = np.array([scale_factor, scale_factor, scale_factor])

    def _morph_chimerax_fired(self):
        # Morph maps in chimerax based on different ordering methods
        idm = np.asarray(list(self.selections.values()))
        sel_names = list(self.selections.keys())
        if len(self.selections) > 1:
            if self.morphing_choice == "Salesman":
                print("Salesman chosen")
                morph_chimerax = FlexMorphChimeraX(self.z_space[idm], sel_names, self.mode, self.path, **self.class_inputs)
                morph_chimerax.showSalesMan()
            elif self.morphing_choice == "Random walk":
                print("Random walk chosen")
                morph_chimerax = FlexMorphChimeraX(self.z_space[idm], sel_names, self.mode, self.path,
                                                   **self.class_inputs)
                morph_chimerax.showRandomWalk()

    def picker_callback(self, picker):
        """ Picker callback: this get called when on pick events.
        """
        if picker.actors.last_actor in self.ipw_sel.actor.actors:
            # Find which data point corresponds to the point picked:
            # we have to account for the fact that each data point is
            # represented by a glyph with several points
            glyph_points = self.ipw_sel.glyph.glyph_source.glyph_source.output.points.to_array()
            point_id = int(picker.point_id / glyph_points.shape[0])
            # If the no points have been selected, we have '-1'
            if point_id != -1:
                # Retrieve the coordinates coorresponding to that data
                # point
                self.current_sel = list(self.selections.keys())[point_id]
                px, py, pz = self.ipw_sel.mlab_source.x[point_id], \
                             self.ipw_sel.mlab_source.y[point_id], \
                             self.ipw_sel.mlab_source.z[point_id]

                self.moveViewToPosition([px, py, pz], None)

                # Update label
                self.ipw_label.trait_set(position=[px + 5., py + 5., pz + 5.], text=self.current_sel)
                self.ipw_label.actor.property.color = (1, 1, 1)

    # ---------------------------------------------------------------------------
    # Conversion functions
    # ---------------------------------------------------------------------------
    def interpolateGrid(self):
        mesh = pv.PolyData(self.data)
        mesh["interp_val"] = self.interp_val
        mesh.set_active_scalars("interp_val")
        grid = pv.create_grid(mesh, dimensions=(300, 300, 300))
        self.origin = grid.origin
        self.sr = grid.spacing
        interpolated = grid.interpolate(mesh, radius=0.1, sharpness=1)
        self.grid = interpolated.active_scalars.reshape((300, 300, 300))

    def moveDataToGrid(self):
        self.data[:, 0] = (self.data[:, 0] - self.origin[0]) / self.sr[0]
        self.data[:, 1] = (self.data[:, 1] - self.origin[1]) / self.sr[1]
        self.data[:, 2] = (self.data[:, 2] - self.origin[2]) / self.sr[2]

    def getDataAtThreshold(self, data, level):
        return np.asarray(np.where(data >= level))

    # ---------------------------------------------------------------------------
    # Read functions
    # ---------------------------------------------------------------------------
    def readMap(self, file):
        map = ImageHandler().read(file).getData()
        return map

    def readSelectionsDict(self):
        pathFile = os.path.join(self.path, "selections_dict.pkl")
        selections_dict_file = open(pathFile, "rb")
        self.selections = pickle.load(selections_dict_file)

    # ---------------------------------------------------------------------------
    # Write functions
    # ---------------------------------------------------------------------------
    def writeVectorFile(self, vector):
        pathFile = os.path.join(self.path, self.vector_file)
        with open(pathFile, 'w') as fid:
            fid.write(' '.join(map(str, vector.reshape(-1))) + "\n")

    def saveSelectionsDict(self):
        pathFile = os.path.join(self.path, "selections_dict.pkl")
        selections_dict_file = open(pathFile, "wb")
        pickle.dump(self.selections, selections_dict_file)
        selections_dict_file.close()

    # ---------------------------------------------------------------------------
    # View functions
    # ---------------------------------------------------------------------------
    def moveViewToPosition(self, position, axis_name):
        setattr(self, 'cursor_position', np.asarray(position))
        for other_axis, axis_number in self._axis_names.items():
            if other_axis == axis_name:
                continue
            ipw3d = getattr(self, 'ipw_3d_%s' % other_axis)
            ipw3d.ipw.slice_position = position[axis_number]

        # Update 3D line position
        for axis, normal in zip(["x", "y", "z"], np.eye(3)):
            line3d = getattr(self, 'line_3d_%s' % axis)
            grid_size = np.asarray(self.grid.shape)
            position_np = np.asarray(position)
            point1 = position_np * (1 - normal)
            point2 = grid_size * normal + point1
            x = [point1[0], point2[0]]
            y = [point1[1], point2[1]]
            z = [point1[2], point2[2]]
            line3d.mlab_source.reset(x=x, y=y, z=z)

        # Update real time conformation
            pos = np.asarray([position[0], position[1], position[2]]).reshape(1, -1)
            _, ind = self.kdtree_data.query(pos, k=1)

            if self.generate_map is not None:
                if self.mode == "Zernike3D":
                    self.generate_map("reference.mrc", "mask.mrc", "deformed.mrc",
                                      self.path, self.z_space[ind[0], :],
                                      int(self.class_inputs["L1"]), int(self.class_inputs["L2"]), 32)
                    self.generated_map = self.readMap(os.path.join(self.path, "deformed.mrc"))
                elif self.mode == "CryoDrgn":
                    self.generate_map(self.z_space[ind[0], :], self.class_inputs["weights"],
                                      self.class_inputs["config"], self.path,
                                      downsample=int(self.class_inputs["boxsize"]),
                                      apix=float(self.class_inputs["sr"]))
                    self.generated_map = self.readMap(os.path.join(self.path, "vol_000.mrc"))
                elif self.mode == "HetSIREN":
                    self.generate_map(self.class_inputs["weights"], self.z_space[ind[0], :],
                                      self.path, step=self.class_inputs["step"])
                    self.generated_map = self.readMap(os.path.join(self.path, "decoded_map_class_01.mrc"))
                elif self.mode == "NMA":
                    self.generate_map(self.class_inputs["weights"], self.z_space[ind[0], :],
                                      self.path, sr=self.class_inputs["sr"])
                    self.generated_map = self.readMap(os.path.join(self.path, "decoded_map_class_01.mrc"))

        volume = getattr(self, 'ipw_map')
        val_max = np.amax(volume.mlab_source.m_data.scalar_data)
        val_min = np.amin(volume.mlab_source.m_data.scalar_data)
        volume.mlab_source.reset(scalars=self.generated_map)
        if val_max == val_min:
            volume.contour.contours = [0.0001]

    # ---------------------------------------------------------------------------
    # The layout of the dialog created
    # ---------------------------------------------------------------------------
    view = View(HSplit(
        HGroup(
            Group(
                Item('scene_y',
                     editor=SceneEditor(scene_class=Scene),
                     height=250, width=300),
                Item('scene_z',
                     editor=SceneEditor(scene_class=Scene),
                     height=250, width=300),
                show_labels=False,
            ),
            Group(
                Item('scene_x',
                     editor=SceneEditor(scene_class=Scene),
                     height=250, width=300),
                Item('scene3d',
                     editor=SceneEditor(scene_class=MayaviScene),
                     height=250, width=300),
                show_labels=False,
            ),
            VGroup(
                Group(
                    Group(
                        Item('show_cut_planes'),
                        # Item('show_selections'),
                        columns=2
                    ),
                    Item('opacity',
                         editor=RangeEditor(format='%.02f', low_name='op_min', high_name='op_max', mode='slider')
                         ),
                    Item('threshold',
                         editor=RangeEditor(format='%.02f', low_name='op_min', high_name='op_max', mode='slider')
                         ),
                    Group(
                        Item('n_clusters', show_label=False, editor=TextEditor()),
                        Item('compute_kmeans'),
                        show_labels=False, columns=2
                    ),
                    Group(
                        Item('scale_factor', show_label=False, editor=TextEditor()),
                        Item('scale_pts'),
                        show_labels=False, columns=2
                    ),
                    Group(
                        Item('save_selections'),
                        Item('add_selection'),
                        Item('remove_selection'),
                        Item('selection_name', show_label=False, editor=TextEditor()),
                        Item('rename_selection'),
                        show_labels=False, columns=3
                    ),
                    Group(
                        Item('morphing_choice',
                             editor=EnumEditor(
                                 values={"Salesman": "Salesman",
                                         "Random walk": "Random walk"}
                             )),
                        Item('morph_chimerax', style='custom', show_label=False),
                        columns=2)
                ),
                Group(
                    Item('scene_c',
                         editor=SceneEditor(scene_class=MayaviScene),
                         height=250, width=300),
                    show_labels=False,
                )
            ),
        ),
    ),
        resizable=True,
        title='Volume Slicer',
        icon=os.path.join(os.path.dirname(flexutils.__file__), "icon_square.png")
    )

if __name__ == '__main__':
    import argparse

    # Input parameters
    parser = argparse.ArgumentParser()
    parser.add_argument('--data', type=str, required=True)
    parser.add_argument('--z_space', type=str, required=True)
    parser.add_argument('--interp_val', type=str, required=True)
    parser.add_argument('--path', type=str, required=True)
    parser.add_argument('--mode', type=str, required=True)

    def float_or_str(s):
        try:
            return float(s)
        except ValueError:
            return s

    parsed, unknown = parser.parse_known_args()
    for arg in unknown:
        if arg.startswith(("-", "--")):
            # you can pass any arguments to add_argument
            parser.add_argument(arg.split('=')[0], type=float_or_str)

    args = parser.parse_args()

    # Read and generate data
    data = np.loadtxt(args.data)
    z_space = np.loadtxt(args.z_space)
    interp_val = np.loadtxt(args.interp_val)

    # Input
    input_dict = vars(args)
    input_dict['data'] = data
    input_dict['z_space'] = z_space
    input_dict['interp_val'] = interp_val

    # Initialize volume slicer
    m = VolumeSlicer(**input_dict)
    m.configure_traits()



# **************************************************************************
# *
# * Authors:  David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import os

from pyworkflow.object import Set
from pyworkflow.utils import removeExt
from pyworkflow.viewer import DESKTOP_TKINTER, WEB_DJANGO, ProtocolViewer
import pyworkflow.protocol.params as params

from pwem.viewers import ChimeraView

from flexutils.protocols.xmipp.protocol_apply_field_nma import XmippApplyFieldNMA


class XmippApplyFieldNMAView(ProtocolViewer):
    """ Visualize a NMA structure """
    _label = 'viewer apply field NMA'
    _targets = [XmippApplyFieldNMA]
    _environments = [DESKTOP_TKINTER, WEB_DJANGO]
    OPEN_FILE = "open %s\n"
    VIEW = "view\n"

    def __init__(self, **kwargs):
        ProtocolViewer.__init__(self, **kwargs)
        self.only_apply = True
        self.deformed = self.protocol.deformed
        self.have_set = isinstance(self.deformed, Set)
        self.choices = list(self.deformed.getIdSet())

    def _defineParams(self, form):
        form.addSection(label='Show deformation')
        form.addParam('idChoice', params.EnumParam,
                      condition='have_set',
                      choices=[str(idx) for idx in self.choices], default=0,
                      label='Structure to display', display=params.EnumParam.DISPLAY_COMBO,
                      help='Select which strucutre to display from the IDs of the set')
        form.addParam('doShowStrain', params.LabelParam,
                      label="Display the strain deformation")
        form.addParam('doShowRotation', params.LabelParam,
                      label="Display the rotation deformation")
        form.addParam('doShowPDB', params.LabelParam,
                      label="Display original and deformed PDB or volume")
        form.addParam('doShowMorph', params.LabelParam,
                      label="Display a morphing between the original and deformed PDB or volume")

    def _getVisualizeDict(self):
        if self.have_set:
            self.chosen = self.deformed[self.choices[self.idChoice.get()]]
        else:
            self.chosen = self.deformed

        return {'doShowStrain': self._doShowStrainStruct,
                'doShowRotation': self._doShowRotationStruct,
                'doShowPDB': self._doShowPDB,
                'doShowMorph': self._doShowMorph}

    # ------------------- Mode Structure Methods -------------------
    def _doShowStrainStruct(self, param=None):
        scriptFile = self.protocol._getPath('strain_chimera.cxc')
        fhCmd = open(scriptFile, 'w')

        fnbase = removeExt(self.chosen.getFileName())
        fnStrain = os.path.abspath(fnbase)

        fhCmd.write(self.OPEN_FILE % (fnStrain + "_strain.pdb"))
        fhCmd.write("show cartoons\n")
        fhCmd.write("cartoon style width 1.5 thick 1.5\n")
        fhCmd.write("style stick\n")
        fhCmd.write('color by occupancy palette rainbow\n')
        fhCmd.write(self.VIEW)
        fhCmd.close()

        view = ChimeraView(scriptFile)
        return [view]

    def _doShowRotationStruct(self, param=None):
        scriptFile = self.protocol._getPath('strain_chimera.cxc')
        fhCmd = open(scriptFile, 'w')

        fnbase = removeExt(self.chosen.getFileName())
        fnStrain = os.path.abspath(fnbase)

        fhCmd.write(self.OPEN_FILE % (fnStrain + "_rotation.pdb"))
        fhCmd.write("show cartoons\n")
        fhCmd.write("cartoon style width 1.5 thick 1.5\n")
        fhCmd.write("style stick\n")
        fhCmd.write('color by occupancy palette rainbow\n')
        fhCmd.write(self.VIEW)
        fhCmd.close()

        view = ChimeraView(scriptFile)
        return [view]

    def _doShowPDB(self, obj, **kwargs):
        scriptFile = self.protocol._getPath('pdb_deform_chimera.cxc')
        fhCmd = open(scriptFile, 'w')
        inputFile = self.protocol.inputStruct.get().getFirstItem().getFlexInfo().refStruct.get()
        inputFile = os.path.abspath(inputFile)
        outputFile = os.path.abspath(self.chosen.getFileName())

        fhCmd.write(self.OPEN_FILE % inputFile)
        fhCmd.write(self.OPEN_FILE % outputFile)
        # fhCmd.write("start Model Panel\n")
        fhCmd.write("show cartoons\n")
        fhCmd.write("cartoon style width 1.5 thick 1.5\n")
        fhCmd.write("style stick\n")
        fhCmd.write("color bymodel\n")
        fhCmd.close()

        view = ChimeraView(scriptFile)
        return [view]

    def _doShowMorph(self, obj, **kwargs):
        scriptFile = self.protocol._getPath('pdb_deform_chimera.cxc')
        fhCmd = open(scriptFile, 'w')
        inputFile = self.protocol.inputStruct.get().getFirstItem().getFlexInfo().refStruct.get()
        inputFile = os.path.abspath(inputFile)
        outputFile = os.path.abspath(self.chosen.getFileName())

        fhCmd.write(self.OPEN_FILE % inputFile)
        fhCmd.write(self.OPEN_FILE % outputFile)
        fhCmd.write("hide models\n")
        fhCmd.write("morph #1,2 frames 50 play false\n")
        fhCmd.write("coordset #3 1,\n")
        fhCmd.write("wait 50\n")
        fhCmd.write("coordset #3 50,1\n")
        fhCmd.close()

        view = ChimeraView(scriptFile)
        return [view]
    # ------------------- ------------------- -------------------




# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import numpy as np
from xmipp_metadata.image_handler import ImageHandler

from pyworkflow import NEW
from pyworkflow.protocol.params import PointerParam, FloatParam, EnumParam
import pyworkflow.utils as pwutils
from pyworkflow.object import Boolean

from pwem.viewers.showj import *
from pwem.protocols import ProtAnalysis3D
from pwem.objects import SetOfVolumes, Volume, SetOfParticles

from flexutils.viewers.imagej_viewers.viewer_ij import launchIJForSelection
from flexutils.utils import getOutputSuffix
import flexutils.constants as const
import flexutils

from xmipp3.convert import geometryFromMatrix
import xmipp3


class ProtFlexSelectViews(ProtAnalysis3D):
    """ Compare different (rot,tilt) views of different maps and interactively select regions to filter/score
     a SetOfParticles """

    _label = 'select views'
    _devStatus = NEW
    OUTPUT_PREFIX = 'selectedParticles'

    # --------------------------- DEFINE param functions ----------------------
    def _defineParams(self, form):
        form.addSection(label='General parameters')
        form.addParam('reference', PointerParam, label="Reference map",
                      pointerClass='Volume', important=True,
                      help='Map to be used as reference for the comparisons')
        form.addParam('compareMaps', PointerParam, label="Maps to compare to",
                      pointerClass='SetOfVolumes, Volume', important=True,
                      help='Maps to compare the reference views to at different (rot,tilt) combinations')
        form.addParam('particles', PointerParam, label="Particles to filter/score",
                      pointerClass='SetOfParticles', important=True)
        form.addParam('mode', EnumParam, choices=['score', 'filter'],
                      default=0, display=EnumParam.DISPLAY_HLIST,
                      label="Score or filter particles?",
                      help="\t * Score: all particles will be added to the output, but they will be "
                           "enabled/disabled based on the selection results \n"
                           "\t * Filter: only particles belonging to the selections will be added "
                           "to the output")
        form.addParam('targetResolution', FloatParam, label="Target resolution (A)", default=8.0,
                      help="In Angstroms, the images and the volume are rescaled so that this resolution is at "
                           "2/3 of the Fourier spectrum.")
        form.addParam('angSampling', FloatParam, default=5.0, label="Angular sampling",
                      help="Angular step determining how many combinations of (rot,tilt) will be "
                           "evaluated during the comparison")
        form.addParallelSection(threads=4, mpi=0)

    # --------------------------- INSERT steps functions ----------------------
    def _insertAllSteps(self):
        self._insertFunctionStep(self.computeCorrImageStep)
        self._insertFunctionStep(self.launchIJGUIStep, interactive=True)

    def _createOutput(self):
        particles = self.particles.get()
        mode = self.mode.get()
        angSampling = self.angSampling.get()
        outFile = self._getExtraPath("combined_corrImage.txt")

        # Move maps to tmp path and compute corr image
        self.newAngSampling = 360. / np.round(360. / angSampling)

        # Define polygons based on selected borders
        polygons_file = self._getExtraPath("polygons.dat")
        args = "--input %s --angs %f --output %s" % (outFile, self.newAngSampling, polygons_file)
        program = "polygon_from_vertexes.py"
        program = flexutils.Plugin.getProgram(program)
        self.runJob(program, args)

        # Create output object
        suffix = getOutputSuffix(self, SetOfParticles)
        output_particles = self._createSetOfParticles(suffix)
        output_particles.copyInfo(particles)
        output_particles.setHasCTF(particles.hasCTF())

        # Loop particles and determine if the lie within the polygon delimited areas
        points_file = self._getExtraPath("point.txt")
        decision_file = self._getExtraPath("decision.txt")
        angles_vec = []
        for particle in particles.iterItems(iterate=False):
            _, angles = geometryFromMatrix(particle.getTransform().getMatrix(), True)
            rot = angles[0]
            tilt = angles[1]
            # Make sure angles are positive
            rot = rot if rot >= 0. else rot + 360.
            tilt = tilt if tilt >= 0. else tilt + 360.
            angles_vec.append([tilt, rot])
        np.savetxt(points_file, np.asarray(angles_vec))

        args = "--input %s --polygons %s --output %s" % (points_file, polygons_file, decision_file)
        program = "check_inside_roi.py"
        program = flexutils.Plugin.getProgram(program)
        self.runJob(program, args)

        # Loop particles and filter/score them according to polygon delimited areas
        in_area_vec = np.loadtxt(decision_file)
        idx = 0
        for particle in particles.iterItems(iterate=False):
            in_area = in_area_vec[idx]
            aux_particle = particle.clone()
            if in_area == 1:
                output_particles.append(aux_particle)
            elif in_area == 0 and mode == 0:
                aux_particle.setEnabled(False)
                output_particles.append(aux_particle)
            idx += 1

        # Save new output
        name = self.OUTPUT_PREFIX + suffix
        args = {}
        args[name] = output_particles
        self._defineOutputs(**args)
        self._defineSourceRelation(particles, output_particles)
        self._updateOutputSet(name, output_particles, state=output_particles.STREAM_CLOSED)

    # --------------------------- STEPS functions -----------------------------
    def computeCorrImageStep(self):
        reference = self.reference.get()
        compareMaps = self.compareMaps.get()
        angSampling = self.angSampling.get()

        # Compute new dimensions
        newTs = self.getNewTs()
        newXDim = self.getNewDim(newTs)

        # Move reference to tmp path
        ih = ImageHandler()
        Xdim = self.reference.get().getDim()[0]
        inputFile = reference.getFileName()
        refFile = self._getTmpPath('reference.mrc')
        if pwutils.getExt(inputFile) == ".mrc":
            inputFile += ":mrc"
        if Xdim != newXDim:
            self.runJob("xmipp_image_resize",
                        "-i %s -o %s --dim %d " % (inputFile,
                                                   refFile,
                                                   newXDim),
                        numberOfMpi=1, env=xmipp3.Plugin.getEnviron())
        else:
            ih.convert(inputFile, refFile)

        # Move maps to tmp path and compute corr image
        self.newAngSampling = 360. / np.round(360. / angSampling)
        size_rot = int(360. / self.newAngSampling)
        size_tlt = int(180. / self.newAngSampling)
        combined_corr_image = np.ones([size_rot + 1, size_tlt + 1])

        if isinstance(compareMaps, SetOfVolumes):
            iterator = compareMaps.iterItems()
        elif isinstance(compareMaps, Volume):
            iterator = [compareMaps]

        for idx, compareMap in enumerate(iterator):
            inputFile = compareMap.getFileName()
            mapFile = self._getTmpPath('map_%d.mrc' % (idx + 1))
            if pwutils.getExt(inputFile) == ".mrc":
                inputFile += ":mrc"
            corrImageFile = self._getExtraPath('corrImage_%d.mrc' % (idx + 1))
            if Xdim != newXDim:
                self.runJob("xmipp_image_resize",
                            "-i %s -o %s --dim %d " % (inputFile,
                                                       mapFile,
                                                       newXDim),
                            numberOfMpi=1, env=xmipp3.Plugin.getEnviron())
            else:
                ih = ImageHandler()
                ih.convert(inputFile, mapFile)

            # Compute corr image
            self.runJob("xmipp_compare_views",
                        "-v1 %s -v2 %s -o %s --degstep %f --thr %d "
                        % (refFile, mapFile, corrImageFile, self.newAngSampling, self.numberOfThreads.get()),
                        env=xmipp3.Plugin.getEnviron())

            # Combine all corr images
            corr_image = ih.read(corrImageFile).getData()
            combined_corr_image *= corr_image

        # Saved combined corr image
        ih.write(combined_corr_image, self._getExtraPath("combined_corrImage.mrc"))


    def launchIJGUIStep(self):
        corrImageFile = self._getExtraPath("combined_corrImage.mrc")

        path = self._getExtraPath()
        launchIJForSelection(path, corrImageFile)

        if os.path.isfile(self._getExtraPath("combined_corrImage.txt")):
            self._createOutput()

    # --------------------------- UTILS functions ----------------------------
    def getNewDim(self, Ts):
        reference = self.reference.get()
        Xdim = reference.getXDim()
        return int(Xdim * reference.getSamplingRate() / Ts)

    def getNewTs(self):
        reference = self.reference.get()
        targetResolution = self.targetResolution.get()
        Ts = reference.getSamplingRate()
        newTs = targetResolution / 3.0
        newTs = max(Ts, newTs)
        return newTs

    # --------------------------- INFO functions -----------------------------
    def _summary(self):
        summary = []
        if self.getOutputsSize() >= 1:
            for _, outRegions in self.iterOutputAttributes():
                summary.append("Output *%s*:" % outRegions.getNameId().split('.')[1])
                summary.append("    * Total # particles: *%s*" % outRegions.getSize())
        else:
            summary.append("Output ROIs not ready yet.")
        return summary

    def _methods(self):
        return [
            "Interactive selection of best views for flexibility analysis with ImageJ and "
            "filtering/scoring based on selections",
        ]



# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import os

import numpy as np
from pwem.protocols import ProtAnalysis3D
from pwem.objects import Volume

import pyworkflow.protocol.params as params
import pyworkflow.utils as pwutils
from pyworkflow.object import Float, Integer, String

import flexutils
from flexutils.utils import readZernikeFile, getXmippFileName
import flexutils.constants as const


class XmippMatchDeformMapZernike3D(ProtAnalysis3D):
    """ Zernike3D deformation field computation between two map based on voxel matches"""
    _label = 'map match and deform - Zernike3D'
    match_method = ["skeleton", "border"]

    # --------------------------- DEFINE param functions --------------------------------------------
    def _defineParams(self, form):
        form.addSection(label='Input')
        form.addParam('input', params.PointerParam, label="Input map",
                      pointerClass='Volume',
                      help="Map to be deformed")
        form.addParam('reference', params.PointerParam, label="Reference map",
                      pointerClass='Volume',
                      help="Target map")
        form.addParam('l1', params.IntParam, default=7,
                      label='Zernike Degree',
                      expertLevel=params.LEVEL_ADVANCED,
                      help='Degree Zernike Polynomials of the deformation=1,2,3,...')
        form.addParam('l2', params.IntParam, default=7,
                      label='Harmonical Degree',
                      expertLevel=params.LEVEL_ADVANCED,
                      help='Degree Spherical Harmonics of the deformation=1,2,3,...')
        form.addParam('matchMode', params.EnumParam, choices=self.match_method, default=0,
                      label='Voxel matching algorithm', display=params.EnumParam.DISPLAY_HLIST,
                      expertLevel=params.LEVEL_ADVANCED,
                      help='- Skeleton mode: faster and works well on maps presenting similar conformations.\n'
                           '- Border mode: slower but able to find matches when conformations are not similar')
        form.addParam('gsSteps', params.IntParam,
                      allowsNull=True,
                      label='Global registration steps',
                      expertLevel=params.LEVEL_ADVANCED,
                      help='Perform a global registration of the maps before computing the deformation fields')

    # --------------------------- INSERT steps functions -----------------------
    def _insertAllSteps(self):
        self._insertFunctionStep("deformStep")
        self._insertFunctionStep("createOutputStep")

    # --------------------------- STEPS functions ------------------------------
    def deformStep(self):
        input = getXmippFileName(self.input.get().getFileName())
        reference = getXmippFileName(self.reference.get().getFileName())
        output = self._getExtraPath("map_deformed.mrc")
        l1 = self.l1.get()
        l2 = self.l2.get()
        args = "--i %s --r %s --o %s --l1 %d --l2 %d --%s" \
               % (input, reference, output, l1, l2, self.match_method[self.matchMode.get()])
        if self.gsSteps.get():
            args += " --gs %d" % self.gsSteps.get()
        program = "find_z_clnm_map.py"
        program = flexutils.Plugin.getProgram(program)
        self.runJob(program, args)

    def createOutputStep(self):
        z_clnm_file = self._getExtraPath("z_clnm.txt")
        basis_params, z_clnm = readZernikeFile(z_clnm_file)
        rmsd_def = np.loadtxt(self._getExtraPath("rmsd_def.txt"))

        L1 = Integer(self.l1.get())
        L2 = Integer(self.l2.get())
        Rmax = Float(basis_params[2])
        deformation = Float(rmsd_def[2])
        # mask = String(self._getExtraPath("mask.mrc"))

        outFile = self._getExtraPath("map_deformed.mrc")
        vol = Volume()
        vol.setLocation(outFile)
        vol.setSamplingRate(self.input.get().getSamplingRate())
        vol.L1 = L1
        vol.L2 = L2
        vol.Rmax = Rmax
        vol._xmipp_sphDeformation = deformation
        vol._xmipp_sphCoefficients = String(','.join(['%f' % c for c in z_clnm[0]]))
        # vol.refMask = mask
        self._defineOutputs(deformedMap=vol)
        self._defineSourceRelation(self.input, vol)
        self._defineSourceRelation(self.reference, vol)

    # --------------------------- UTILS functions ------------------------------
